#include "Latticeoids3d.h"

/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 */

//This is a constructor that populates the BeOids object with molecules, with the configuration set by cnsts->latticetype.
//Currently, can only build simple cubic lattices or read in a coordinate file.
//Inputs:
//	BeOids - the entire ensemble of molecules (passthrough)
//	s - a type A molecule to be copied on to the lattice
//	t - a type B molecule to be copied on to the lattice
LatticeOids3d::LatticeOids3d(Molecule** BeOids,Molecule &s,Molecule &t)
{
	Cnsts *cnsts = GetCnsts();
	my_BeOids = BeOids;
	
	//Create array of molecules -- BeOids. This is currently hard-coded for two types of molecules, and will need to be
	//expanded to handle an arbitrary number of types at a later date. Can be done with either a sequential (load all As, then all Bs)
	//or a random boxloader.
	if (cnsts->loader == 1) SeqLoad(BeOids, s, t);
	else if (cnsts->loader == 2) RandLoad(BeOids, s, t);
	else if (cnsts->loader == 3)
	{
		string cfn = cnsts->filein + int2str(cnsts->simnum) + "chk" + cnsts->chkpt + "coords.dat";
		LoadCoordFile(BeOids, s, t,cfn);
		printf("Coordinate file %s sucessfully read.\n", cfn.c_str());
		return;
	}
	else
	{
		printf("Invalid boxloader specified.\n");
		exit(1);
	}

	//Construct simple cubic lattice. All current simulation types start from this lattice.
	SimpleCubic(BeOids);
	
	int i=0;

	//Randomize orientations
	for (i=0;i<cnsts->n_oids;i++)
	{
		for (int j=0;j<5;j++)
		{
			BeOids[i]->Rotate();
		}
	}

	//Randomize positions. Using this routine or latticetype == 2 is not essential, as randomizesteps handles
	//most of the work for breaking up the lattice. However, randomizesteps incorporates vdw interactions, so
	//if you want to start with a less energetically favored spatial distribution of molecules, use latticetype 2, and
	//if you want the vdw interactions to dictate the starting configuration, use randomizesteps. Do not modify
	//this routine for other latticetypes without making sure you are calling the correct Translate function.
	if (cnsts->latticetype==2)
	{
		if ((cnsts->rot_notransA==1)||(cnsts->rot_notransB==1))
		{
			printf("Translating particles that shouldn't be able to move is not recommended.\n");
			exit(1);
		}
		

		for (int k=0;k<10;k++)
		{
			for (i=0;i<cnsts->n_oids;i++)
			{
				for (int j=0;j<3;j++)
				{
					BeOids[i]->Translate();
				}
			}
		}
	}
}

//Destructor for LatticeOids3D. While it looks tempting to replace this loop with a delete[], it will
//result in a segfault if you run a batch job. - LEJ
LatticeOids3d::~LatticeOids3d()
{
	Cnsts *cnsts = GetCnsts();
	for(int j = 0; j < cnsts->n_oids; j++) delete my_BeOids[j];
}

//Simple Cubic Lattice generator. Assigns coordinates to a pre-constructed ensemble of molecules to form a simple cubic lattice.
void SimpleCubic(Molecule** BeOids)
{
	Cnsts *cnsts = GetCnsts();
	double x, y, z;
	int i = 0;
	for (int xi=0;xi<cnsts->lin_dim[0];xi++)
	{
		for (int yi=0;yi<cnsts->lin_dim[1];yi++)
		{
			for (int zi=0;zi<cnsts->lin_dim[2];zi++)
			{
				
				x=(double)xi - cnsts->lin_dim[0]/(2.0) + 0.5;
				y=(double)yi - cnsts->lin_dim[1]/(2.0) + 0.5;
				z=(double)zi - cnsts->lin_dim[2]/(2.0) + 0.5;
				BeOids[i]->center.vec[0] = cnsts->nndist*(x);
				BeOids[i]->center.vec[1] = cnsts->nndist*(y);
				BeOids[i]->center.vec[2] = cnsts->nndist*(z);
				i++;
			}
		}
	}
}

//Sequential loader. Should only be used for homogeneous systems or systems where you *really* want
//to start with two distinct layers, as it loads all A oids, then all B oids. Called if cnsts->loader == 1
void SeqLoad(Molecule** BeOids, Molecule &s, Molecule &t)
{
	Cnsts *cnsts = GetCnsts();
	int i;
	for(i=0;i<cnsts->n_oids;i++)
	{
		if (i< cnsts->num_A)
		{
			BeOids[i] = new Molecule(s);
			BeOids[i]->mytype = 0;
			//BeOids[i]->n_type = cnsts->num_A;
		}
		if (i >= cnsts->num_A)
		{
            BeOids[i] = new Molecule(t);
			BeOids[i]->mytype = 1;
			//BeOids[i]->n_type = cnsts->num_B;
		}
		BeOids[i]->myindex = i;
		BeOids[i]->LastV = 0;
	}
}

//Random loader for binary mixtures; makes sure all of the As aren't in one place and the Bs in another
//Used with loader 2
void RandLoad(Molecule** BeOids, Molecule &s, Molecule &t)
{
	Cnsts *cnsts = GetCnsts();
	int i = 0;
	int nA = 0, nB = 0;	//Start with zero molecules placed
	double fA = (double)cnsts->num_A/(double)(cnsts->num_A + cnsts->num_B);	//Fraction of type A
	double r;
	
	while (i < cnsts->n_oids)
	{
		r = ran2();
		
		if (r < fA && nA < cnsts->num_A)	//If out of type B, will create an A regardless of r
		{
			BeOids[i] = new Molecule(s);
			BeOids[i]->mytype = 0;
			BeOids[i]->myindex = i;
			BeOids[i]->LastV = 0.0;
			i++;
			nA++;	//Remove an A from the inventory. Will no longer place As once all are placed.
		}

		if (r > fA && nB < cnsts->num_B)	//If out of type A, will create a B regardless of r
		{
            BeOids[i] = new Molecule(t);
			BeOids[i]->mytype = 1;
			BeOids[i]->myindex = i;
			BeOids[i]->LastV = 0.0;
			i++;
			nB++;	//Remove a B from the inventory. Will no longer place Bs once all are placed.
		}
	}
}


//Parser and loader for coordinate files. Used with loader 3
void LoadCoordFile(Molecule** BeOids,Molecule &s,Molecule &t, string cfn)
{
	
	Cnsts *cnsts = GetCnsts();
	
	//Allocate memory
	int i,j,n,num,type;
	
	//Open stream
	fstream cfile;
	cfile.open(cfn.c_str(),fstream::in);
	
		//Check stream integrity
	if (cfile.good() != 1)
	{
		printf("Error opening file '%s'\n", cfn.c_str());
		exit(4);
	}
	
	//Skip over header information
	advanceline(cfile, 6);
	
	//Iterate over N molecules
	for (n = 0; n < cnsts->n_oids; n++)
	{
		//Get index and type
		num = gettdint(cfile);
		if (num != n)
		{
			printf("Invalid particle numbering scheme! Expected particle %i, found particle %i.\n", n, num);
			exit(4);
		}
		type = gettdint(cfile);
		
		//Create molecule
		if (type == 0) BeOids[n] = new Molecule(s);
		if (type == 1) BeOids[n] = new Molecule(t);
		BeOids[n]->mytype = type;
		BeOids[n]->myindex = n;
		BeOids[n]->LastV = 0.0;
		
		//Get position vector
		for(i = 0; i < 3; i++)
		{
			BeOids[n]->center.vec[i] = gettddoub(cfile);
		}
		
		//Get rotation matrix
		for(i = 0; i < 3; i++)
		{
			for (j = 0; j < 3; j++)
			{
				BeOids[n]->rot.mat[i][j] = gettddoub(cfile);
			}
		}
		
		//Exit if EOF, otherwise continue
		if(cfile.good() != 1)
		{
			cfile.close();
			break;
		}
		advanceline(cfile, 2);
	}

	//Check to make sure loading completed sucessfully
	cfile.close();
	if (n < cnsts->n_oids)
	{
		printf("Only %i out of %i particles loaded!",n,cnsts->n_oids);
		exit(4);
	}
	else if (n > cnsts->n_oids)
	{
		printf("Too many (%i out of %i) particles loaded!",n,cnsts->n_oids);
		exit(4);
	}
}

