#include "Molecule.h"
/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 */

//Cnsts *cnsts = GetCnsts();

//Create a molecule from an ellipse. Don't think this one is currently used.
Molecule::Molecule(const Ellipse& e)
{
 SetEllipse(e); 
}

//Set the properties for an ellipse
void Molecule::SetEllipse(const Ellipse& e)
{
	ellipse = &e;
	rot.M3Eye();
	center.V3Zeros();
	still = -1;
	rot_notrans = -1;
	myindex = -1;
	mytype = -1;
}

//Copy constructor
Molecule::Molecule(Molecule& old)

{
	SetEllipse(*(old.ellipse));
	rot = old.rot;
	center = old.center;
	still = old.still;
	rot_notrans = old.rot_notrans;
	myindex = old.myindex;
	mytype = old.mytype;
	averageme = old.averageme;
}


//New constructor; use this one.
Molecule::Molecule(const Ellipse& ellipse, int Still, int Rot_notrans, int mtype, bool avgme)
{
	SetEllipse(ellipse);
	
	this->still = Still;
	this->rot_notrans = Rot_notrans;
	this->mytype = mtype;
	this->averageme = avgme;
}


//Regular translate method for cubic cell with PBCs in an arbitrary number of directions.
//If a move in a non-PBC'd direction would cause the particle to leave the box, the component
//of the move along the offending axis is ignored, but the components along other axes
//are retained. Return value is the squared distance of the move.
double Molecule::Translate()

{
	Cnsts *cnsts = GetCnsts();	
	double trans, pbcadj, transtemp;
	double moved = 0.0;
	
	for (int i = 0; i < 3; i++)
	{
		//Get a random number
		trans = (cnsts->maxtrans)*(2.0*ran2()-1.0);
		
		//Make a move appropriate to the PBCs along that axis
		if (cnsts->PBCs[i] == 1)
		{
			center.vec[i] += trans;
			pbcadj = center.vec[i]/cnsts->boxlength[i];
			center.vec[i] -= cnsts->boxlength[i]*qround(pbcadj);
			moved += trans*trans;
		}
		else
		{
			transtemp = center.vec[i] + trans;
			if (abs(2*transtemp) < cnsts->boxlength[i])
			{
				center.vec[i] = transtemp;
				moved += trans*trans;
			}
		}
	}
	
	return moved;
}

//Translate method for isolated spherical cell (no PBCs)
//If a move would cause a particle to leave the sphere, the old coords are retained.
//Return value is the squared distance of the move
double Molecule::Translate_Sphere()
{
	double randmov;
	double moved = 0.0;
	Vec3 transtemp(center);
		//Vcpy(transtemp, center);
	
	Cnsts *cnsts = GetCnsts();
	
	//Try a random move
	for (int i = 0; i<3; i++)
	{
		randmov = (cnsts->maxtrans)*(2.0*ran2()-1.0);
		transtemp.vec[i] += randmov;
		moved += randmov*randmov;
	}
	
		//double rtemp = Vnorm3(transtemp);
	double rtemp = transtemp.V3Norm();
	
	//Accept the move only if the particle is still in the sphere
	if(rtemp < cnsts->spherecylr)
	{
			//Vcpy(center, transtemp);
		center = transtemp;
	}
	else moved = 0.0;
	
	return moved;
}

//Translate method for cylindrical cell (PBCs currently only available on z-axis)
//If a move would cause a particle to leave the cylinder, the component of the
//move along the offending axis (r or z) is discarded. Moves along other axes
//are retained. Return value is the squared distance of the move.
double Molecule::Translate_Cylinder()
{
	Cnsts *cnsts = GetCnsts();
	
	double trans1, trans2, trans3, chkr, pbcadj, transtemp;
	double moved = 0.0;
	
	//Get all of the random numbers
	trans1 = (cnsts->maxtrans)*(2.0*ran2()-1.0);
	trans2 = (cnsts->maxtrans)*(2.0*ran2()-1.0);
	trans3 = (cnsts->maxtrans)*(2.0*ran2()-1.0);
	
	//Check to see if the r componentof the move is ok
	chkr = sqrt(pow(trans1+center.vec[0], 2) + pow(trans2+center.vec[1],2));
	if (chkr < cnsts->spherecylr)
	{
		center.vec[0] += trans1;
		center.vec[1] += trans2;
		moved = trans1*trans1+trans2*trans2;
	}
		
	//Make a move appropriate to the PBCs along the z-axis
	if (cnsts->PBCs[2] == 1)
	{
		center.vec[2] += trans3;
		pbcadj = center.vec[2]/cnsts->boxlength[2];
		center.vec[2] -= cnsts->boxlength[2]*qround(pbcadj);
		moved += trans3*trans3;
	}
	else
	{
		transtemp = center.vec[2] + trans3;
		if (abs(2*transtemp) < cnsts->boxlength[2])
		{
			center.vec[2] = transtemp;
			moved += trans3*trans3;
		}
	}
	return moved;
}	


//Rotation method for all simulation types
//Rotates around a single axis, picking at random, limited by cnsts->maxrot.
//Return value is the squared angular displacement of the move.
double Molecule::Rotate()
{

    Cnsts *cnsts = GetCnsts();

	Mat33 Rb, tmp;

	//Random number for picking an axis
	double r1 = ran2();

	//Random number for the move itself
    double theta = cnsts->maxrot*(2.0*ran2()-1.0);
    double co = cos(theta);
	double si = sin(theta);

	//Yaw
	if (r1<0.33)

	{
		Rb.mat[1][1] = co; Rb.mat[1][2] = si;

		Rb.mat[2][1] = -1.0*si; Rb.mat[2][2] = co;
	}

	//Pitch
	if((r1>=0.33)&(r1<0.66))

	{
		Rb.mat[0][0] = co; Rb.mat[0][2] = si;

		Rb.mat[2][0] = -1.0*si; Rb.mat[2][2] = co;
	}

	//Roll
	if (r1>0.66)

	{
		Rb.mat[0][0] = co; Rb.mat[0][1] = si;

		Rb.mat[1][0] = -1*si; Rb.mat[1][1] = co;
	}

	tmp = Rb*rot;
	rot = tmp;
	
	return theta*theta;
}



Molecule::~Molecule(void)
{

}

