#include "OneStep.h"
 
/*
 *  2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 */

//OneStep computes a single step of all of the interactions between one particle (kk from main loop) and every other particle (i) in the system.
//Inputs:
//	BeOids - the arracy containing all molecules in the system
//	mus - the array containing all lab-frame dipole moments
//	kk - the index of the active ellipsoid (ellipsoid being moved)
//	Vmm - passthrough for mu-mu energy (actually an output)
//	VLJ - passthrough for Lennard-Jones energy (actually an output)
//	skipmumu - flag used to control whether mu-mu interactions are calculated
void OneStep(Molecule** BeOids, Vec3* mus, int &kk, double &Vmm, double &VLJ, bool &skipmumu)
{
	Cnsts *cnsts = GetCnsts();
	
	//Calculate the distance array
	Vec3* Rmus = new Vec3[cnsts->n_oids];
	get_distances(BeOids, BeOids[kk]->center, Rmus);
	Vec3 tmpRmu;
	
	//Zero energies before calculation
	Vmm = 0.0;
	VLJ = 0.0;
	
	//Initialize flags checking for presence of dipoles
	bool ihasdipole;
	bool domumu;
	
	//Initialize scalar variables
	double distance, mudist, distance_squared, mudistsq;
	double escut_squared = cnsts->escut*cnsts->escut;

	//Loop over all of the 2-body interactions
	for(int i=0;i<cnsts->n_oids;i++)
	{
		ihasdipole = BeOids[i]->ellipse->hasmu;
		if (i != kk)
		{
			
			//Calculate distances
			distance_squared=Rmus[i]*Rmus[i];
			
			//Bomb out if a zero distance is detected to avoid dividing by zero
			if(distance_squared < 1.0e-6)
			{
				printf("Distance from particle %i to particle %i is (close to) 0! Exiting...\n", kk, i);
				exit(3);
			}
			
			if (distance_squared < escut_squared || cnsts->cut == 0)
			{
				//Only do the expensive square root if it is really needed
				distance = sqrt(distance_squared);
				
				//Mu-mu interactions
				if (skipmumu == 0 && ihasdipole*BeOids[kk]->ellipse->hasmu == 1) domumu = 1;
				else domumu = 0;
				
				if (domumu == 1)
				{
					if (cnsts->offctrmu == 0) Vmm += Vmumu(mus[kk], mus[i], Rmus[i], distance, distance_squared);
					
					//Off-center dipole correction - warning: This code is inefficient!
					else
					{
						offcenter_mu(BeOids, tmpRmu, i,kk);
						apply_pbcs(tmpRmu);
						mudist = tmpRmu.V3Norm();
						if(mudist < 1.0e-3)
						{
							printf("Distance from particle %i to particle %i is (close to) 0! Exiting...\n", kk, i);
							exit(3);
						}
						mudistsq = mudist*mudist;
						Vmm += Vmumu(mus[kk], mus[i], tmpRmu, mudist, mudistsq);
					}
				}

					//Vdw interactions (pick whether you want to use the soft-sphere, simple, extended, or Touch LJ function)
					//Invalid choice defaults to simple 6-12 potential.
					//if (cutchk == 1)
				if (distance < cnsts->ljcut || cnsts->cut == 0)
				{
					switch (cnsts->vdwtype)
					{
						case 1:
							VLJ += simpleVLJ_notouch(BeOids,kk,i,distance);
							break;
						case 2:
							VLJ += VLJ_notouch(BeOids,kk,i,distance,distance_squared);
							break;
						case 3:
							VLJ += VSS_notouch(BeOids,kk,i,distance);
							break;
						case 4:
							VLJ += simpleVLJ_touch(BeOids,kk,i,Rmus[i],distance);
							break;
						case 5:
							VLJ += VSE_touch(BeOids,kk,i,Rmus[i],distance);
							break;
						default:
							VLJ += simpleVLJ_notouch(BeOids,kk,i,distance);
					}
				}
			}
		}
	}
	
	//Scale dipole-dipole interactions by background dielectric constant (1 unless implicit polarizability is used)
	Vmm *= cnsts->in2;
	
	//Clean up
	delete[] Rmus;
}


