/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *
 *  TotalV.cpp
 *  Created by ljohnson on 1/7/09.
 *	Total potential energy calculations and helper functions
 */

#include "TotalV.h"

//Total muE energy. This function replaces the one in the old VmuEtotal file, and is kept separate
//because the muE energy needs to be recalculated if E0 is changed during the simulation, as happens if running at constant external potential.
//Inputs:
//	mus - array of all dipole moment vectors in the system
//	Evec - vector containing the effective (internal) poling field E0
//Return value is the total dipole-field energy for the system
double VmuEtotal(Vec3* mus, Vec3 &Evec)
{
	Cnsts *cnsts = GetCnsts();
	
	double VmuE = 0.0;
	
	//Loop over all particles
	for (int i=0;i<cnsts->n_oids;i++)
	{		
		//Calculate the muE energy for each dipole and add to total
		double tmpVmuE = mus[i]*Evec; //Vdot3(mus[i],Evec);
		VmuE -= tmpVmuE;
	}
	
	return VmuE;
}

//Initial potential energy calculation - Provides a reference energy to update as the simulation progresses. Inefficient and N^2, but only called once.
void firststep(Molecule** BeOids, Vec3* mus, Vec3 &Evec, Vec3 &totalVs, bool skipmm)
{
	Cnsts *cnsts = GetCnsts();
	
	double Vmm = 0.0;
	double VLJ = 0.0;
	
	//loop over all N^2 oids. This could later be optimized to (N^2-N)/2, but I wanted to avoid too much code
	//repetition at the moment, so it currently just calls OneStep repeatedly.
	for (int k = 0; k < cnsts->n_oids; k++)
	{
		//MuE energy
		totalVs.vec[0] += VmuE(mus[k], Evec);
		
		//2-Body interactions
		OneStep(BeOids, mus, k, Vmm, VLJ, skipmm);
		totalVs.vec[1] += Vmm;
		totalVs.vec[2] += VLJ;
	}
	
	//Multiply mu-mu and LJ energy by 0.5 to get rid of double-counting. MuE energy no longer halved (bug caught by AT, fixed 05/10/11)
	totalVs.vec[1] *= 0.5;
	totalVs.vec[2] *= 0.5;
	
}


//Function for changing E and T during the run, which evaluates every cycle.
//Changesteps is an array of the steps where changes take place, trackchange
//is a counter that moves forwards one every time a change is made. At minimum,
//this function will set the values before the simulation starts, so all of the
//arrays must have at least one element. This function also controls simulated
//annealing at the beginning of the simulation if it is used.
//Inputs:
//	Evec - effective (local) poling field vector
//	step - current simulation step
//	epsilon - reaction field dielectric constant, used for updating poling field
//Return value is a flag to indicate whether or not dipole-field interactions need to be recalculated
bool setEandT(Vec3 &Evec, long step, double epsilon)
{
	Cnsts* cnsts = GetCnsts();
	
	//VmuE must be recalculated if E changes
	bool recalcuE = 0;

	double e_scale = cnsts->Escale;
	if (cnsts->realfield == 1 && step >= cnsts->randsteps)
	{
		if (cnsts->scrf == 1) e_scale *= (3.0*epsilon)/(2.0*epsilon+cnsts->n2);
		else e_scale *= (2.0*cnsts->epsilon + epsilon)/(2.0*cnsts->epsilon + cnsts->n2);
		Evec.vec[2] = e_scale*(cnsts->Efield);
		recalcuE = 1;
	}
	
	//Simulated annealing for randomization phase
	//This is currently a simple model with the initial temperature dependent on the number of randsteps - LEJ 10/23/2009
	//Updated to use two phases, one in the initial randomization, and one right when electrostatics turn on - LEJ 02/09/2010
	
	//Critical steps
	long cr12 = cnsts->randsteps/2;
	long cr2 = cnsts->randsteps*2;
	
	//Phase 1 - randomization with no electrostatics. Lasts for first half of randsteps.
	if (cnsts->anneal == 1 && step <= cr12)
	{
		//Temperature increment in K.
		double tinc = 0.20;
		
		//Enable annealing, reduce temperature, or disable annealing, depending on step.
		if (step == 0) 
		{
			printf("Using simulated annealing for pre-minimization.\n");
			printf("Initial annealing temp is %fK\nTemperature increment is %fK\n\n",cnsts->T + (double)(cnsts->randsteps/10), tinc);
		}
		
		//Enable annealing and expand move sizes
		else if (step == 1)
		{
			cnsts->kT += cnsts->kB*(double)(cnsts->randsteps/10);
			cnsts->beta = 1.0/cnsts->kT;
			cnsts->maxtrans *= 1.6;
			cnsts->maxrot *= 1.6;
		}
		//Reduce the temperature by tinc K per step during the annealing period
		else if (step < cr12)
		{
			cnsts->kT -= tinc*cnsts->kB;
			cnsts->beta = 1.0/cnsts->kT;
		}
		//Reset move sizes once annealing period is over
		else if (step == cr12)
		{
			cnsts->kT = cnsts->kB*cnsts->T;
			cnsts->beta = 1.0/cnsts->kT;
			cnsts->maxtrans *= 0.625;
			cnsts->maxrot *= 0.625;
			printf("\nSystem has cooled to final temp of %fK\n",cnsts->T);
		}
	}
	
	//Phase 2 - randomization with electrostatics. Lasts for randsteps after randomization phase finishes.
	//Functionally identical to the first simulated annealing phase, but cooling slower for longer
	//For this function, the simulation must not start averaging until at least step 2*randsteps
	else if (cnsts->anneal == 1 && step <= cr2)
	{
		//Temperature increment in K.
		double tinc = 0.10;
		
		//Enable annealing, reduce temperature, or disable annealing, depending on step.
		if (step == cnsts->randsteps) 
		{
			printf("\nUsing simulated annealing on for next %li steps.\n",cr2);
			printf("Initial annealing temp is %fK\nTemperature increment is %fK\n",cnsts->T + (double)(cnsts->randsteps/10), tinc);
		}
		else if (step == (cnsts->randsteps + 1))
		{
			cnsts->kT += cnsts->kB*(double)(cnsts->randsteps/10);
			cnsts->beta = 1.0/cnsts->kT;
			cnsts->maxtrans *= 1.6;
			cnsts->maxrot *= 1.6;
		}
		else if (step > cnsts->randsteps && step < cr2)
		{
			cnsts->kT -= tinc*cnsts->kB;
			cnsts->beta = 1.0/cnsts->kT;
		}
		else if (step == cr2)
		{
			cnsts->kT = cnsts->kB*cnsts->T;
			cnsts->beta = 1.0/cnsts->kT;
			cnsts->maxtrans *= 0.625;
			cnsts->maxrot *= 0.625;
			printf("\nSystem has cooled to final temp of %fK\n",cnsts->T);
		}
	}
	
	//DEBUG - arrays have been converted to single-element variables! BEWARE! -LEJ
	
	//Make change if on a step flagged for a change
	if (step == cnsts->changesteps)
	{

		//Update kT
		double oldkT = cnsts->kT;
		cnsts->kT = (cnsts->kB*cnsts->T);
		cnsts->beta = 1.0/cnsts->kT;
		if (oldkT != cnsts->kT && step != 0)
		{
			printf("T is now %f K\n", cnsts->T);
		}
		
		double oldE = Evec.vec[2];
		if (cnsts->realfield == 0) Evec.vec[2] = cnsts->Escale*cnsts->Efield;
		if (oldE != Evec.vec[2] && step != 0) recalcuE = 1;
		{
			//printf("E0z is now %f MV/m\n", Evec.vec[2]);
			recalcuE = 1;
		}
		
		//Increment change counter
		cnsts->trackchange++;
	}
	
	return recalcuE;
}

//Recalculation of total mu-mu energy only, used for when reaction field dielectric constant is updated
void recalcmm(Molecule** BeOids, Vec3* mus, Vec3 &totalVs)
{
	Cnsts *cnsts = GetCnsts();
	
	//Use square of electrostatic cutoff for speed
	double escut_squared = cnsts->escut*cnsts->escut;
	
	//Initialize distance array
	Vec3* Rmus = new Vec3[cnsts->n_oids];
	
	//Zero energy before calculation
	double Vmm = 0.0;
	
	for (int kk = 0; kk < cnsts->n_oids; kk++)
	{
		//Calculate distances
		get_distances(BeOids, BeOids[kk]->center, Rmus);
	
		//Loop over all of the 2-body interactions
		for(int i=0;i<kk;i++)
		{
		
			//Calculate distances
			double distance_squared = Rmus[i]*Rmus[i];
			
			//Bomb out if a zero distance is detected to avoid dividing by zero
			if(distance_squared < 1.0e-6)
			{
				printf("Distance from oid %i to oid %i is 0! Exiting...\n", kk, i);
				exit(3);
			}
			
			//Mu-mu interactions
			if (distance_squared < escut_squared || cnsts->cut == 0)
			{
				
				double distance = sqrt(distance_squared);
				
				if (cnsts->offctrmu == 0) Vmm += Vmumu(mus[kk], mus[i], Rmus[i], distance, distance_squared);
				
				//Off-center dipole correction - warning: this code is inefficient!
				else
				{
					Vec3 tmpRmu;
					offcenter_mu(BeOids, tmpRmu, i,kk);
					apply_pbcs(tmpRmu);
					double mudist = tmpRmu.V3Norm();
					if(mudist < 1.0e-3)
					{
						printf("Distance from particle %i to particle %i is (close to) 0! Exiting...\n", kk, i);
						exit(3);
					}
					double mudistsq = mudist*mudist;
					Vmm += Vmumu(mus[kk], mus[i], tmpRmu, mudist, mudistsq);
				}
			}
		}
	}
	
	//Clean up
	delete[] Rmus;
	
	//Return updated mu-mu energy
	Vmm *= cnsts->in2;
	totalVs.vec[1] = Vmm;
	
}

