#include "VLJ.h"

/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 */

//Calculation of Lennard-Jones interactions. Notouch variants can only handle spheres, Touch variants handle anisotropy
//using the method of Perram and Wertheim (J. Comp. Phys, 1985). Several functions are available, providing tradeoffs
//between generality and speed
//Variable defitions:
//	BeOids - the entire array of molecules in the simulation
//	kk - the index of the currently molecule
//	i - the index of the molecule it is interacting with
//	distance - the norm of the distance vector between the molecules
//	distance_squared - the square of the distance vector, only needed for generalized LJ potential
//Return value is the LJ energy for one pair of ellipsoids

//Soft sphere - only calculate nuclear repulsion term of LJ potential
double VSS_notouch(Molecule** BeOids, int &kk,int &i,double &distance)
{
	Cnsts *cnsts = GetCnsts();
	
	//Calculate parameters
	double ro = cnsts->pre_sigma[BeOids[i]->mytype][BeOids[kk]->mytype];
	double sx = ro/distance;
	double eps = cnsts->pre_eps[BeOids[i]->mytype][BeOids[kk]->mytype];  //LJ parameter epsilon
	double disp = qpwr(sx,cnsts->LJexp[0]);
	
	//Calculate energy
	double VLJ = eps*cnsts->r*disp*disp;
	return VLJ;
}

//Simple Lennard-Jones potential (nuclear repulsion power is twice that of dispersion, e.g. LJ 12-6 potential)
double simpleVLJ_notouch(Molecule** BeOids, int &kk,int &i,double &distance)
{
	Cnsts *cnsts = GetCnsts();
	
	//Calculate parameters
	double ro = cnsts->pre_sigma[BeOids[i]->mytype][BeOids[kk]->mytype];
	double sx = ro/distance;
	double eps = cnsts->pre_eps[BeOids[i]->mytype][BeOids[kk]->mytype];  //LJ parameter epsilon
	double disp = qpwr(sx,cnsts->LJexp[0]);
	
	//Calculate energy
	double VLJ = eps*(cnsts->r*disp*(disp-1.0));
	return VLJ;
}


//General LJ potential including optional Bruce correction (attenuating Gaussian function in bottom of well)
double VLJ_notouch(Molecule** BeOids, int &kk,int &i,double &distance, double &distance_squared)
{
	Cnsts *cnsts = GetCnsts();

	//Calculate energy
	double ro = cnsts->pre_sigma[BeOids[i]->mytype][BeOids[kk]->mytype];
	double rmrmin = distance - cnsts->Solvent[2];
	double sx = ro/distance;
	double eps = cnsts->pre_eps[BeOids[i]->mytype][BeOids[kk]->mytype];  //LJ parameter epsilon

	//LJ + Bruce correction around r=rmin. Tanh functions have been replaced with a Gaussian for speed
	double VLJ = eps*cnsts->r*(qpwr(sx,cnsts->LJexp[1])-qpwr(sx,cnsts->LJexp[0])) + cnsts->Solvent[1]*exp(cnsts->Solvent[0]*rmrmin*rmrmin);
	
	return VLJ;
}

//Simplified VLJ, but using Touch algorithm to handle non-spherical ellipsoids. This ONLY works for even powers!
double simpleVLJ_touch(Molecule** BeOids, int kk, int i, Vec3 &Rmu, double distance)
{
	Cnsts *cnsts = GetCnsts();
	
	//Determine whether touch should be run or the average radius should be used
	double Rx; //reduced distance (r0/r)^2
	bool touchme = cnsts->pre_touch[BeOids[i]->mytype][BeOids[kk]->mytype];
	if (touchme == 1 && (cnsts->touchtrunc == 0 || distance < cnsts->tcut))
	{
		Rx = touch(BeOids, kk, i, Rmu, distance); //Square of reduced distance corrected based on closest contact
	}
	else
	{
		Rx = cnsts->pre_sigma[BeOids[i]->mytype][BeOids[kk]->mytype]/distance; // (r0/r)
		Rx *= Rx;
	}
	
	//Calculate parameters
	double eps = cnsts->pre_eps[BeOids[i]->mytype][BeOids[kk]->mytype];  //LJ parameter epsilon
	int lj = cnsts->LJexp[0]/2; //Since Rx is already squared, the small term must be halved
	double disp = qpwr(Rx,lj);
	
	//Calculate energy
	double VLJ = eps*(cnsts->r*disp*(disp-1.0));
	return VLJ;
}

//Soft-sphere potential, but using Touch algorithm to handle non-spherical ellipsoids
double VSE_touch(Molecule** BeOids, int kk, int i, Vec3 &Rmu, double distance)
{
	Cnsts *cnsts = GetCnsts();
	
	//Determine whether touch should be run or the average radius should be used
	double Rx; //reduced distance (r0/r)^2
	bool touchme = cnsts->pre_touch[BeOids[i]->mytype][BeOids[kk]->mytype];
	if (touchme == 1 && (cnsts->touchtrunc == 0 || distance < cnsts->tcut))
	{
		Rx = touch(BeOids, kk, i, Rmu, distance); //Reduced distance (r0/r)^2, corrected based on closest contact
	}
	else
	{
		Rx = cnsts->pre_sigma[BeOids[i]->mytype][BeOids[kk]->mytype]/distance; // (r0/r)
		Rx *= Rx;
	}
	//Calculate parameters
	double eps = cnsts->pre_eps[BeOids[i]->mytype][BeOids[kk]->mytype];  //LJ parameter epsilon
	double disp = qpwr(Rx,cnsts->LJexp[0]); //since Rx is already squared, this gives the larger (repulsive term)
	
	//Calculate energy
	double VLJ = eps*cnsts->r*disp;
	return VLJ;
}
