/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *
 *  VecMat.h
 *  Created by ljohnson on 7/24/09. Last updated 02/05/11
 *  Lightweight linear algebra library intended for rapid three-dimensional computation. This library cannot handle
 *	vectors and matrices of any other dimensionality; a more general library should be used if variable or higher
 *	dimensions are required. The library consists of two classes: Vec3 (vectors) and Mat33 (matrices). Vectors are not
 *	tracked as row or column vectors, with behavior instead determined by the function acting on the vector.
 *	Overloaded operators will always assume that vectors are oriented such that their inner dimensions agree.
 *	All double-precision arithmatic operations are overloaded for vector-scalar and matrix-scalar operations, but are not currently
 *	commutative, meaning that the scalar must be placed *AFTER* the vector or matrix in order to avoid an error, as the
 *	opposite arrangement is not yet defined. All legal vector-vector, vector-matrix, and matrix-matrix BLAS operations are overloaded,
 *	and should not have a problem with commutivity. Only the == and != comparison operators are defined for vectors and matrices. All
 *	elementary math functions are inlined and vector-scalar and vector-vector operations were written unrolled. Other functions may be
 *	partially unrolled. The inverse, linear solve, and string conversion functions are not inlined.
 *
 *	IMPORTANT NOTE: Arguments for operators are defined as constant. Expressions of the form B = A*B will result in a segfault.
 *	This may be avoided by using an accumulating operator; e.g. B *= A instead.
 */

#pragma once
#include <cstdio>
#include <ctime>
#include <string>
#include <cmath>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <fstream>

class Mat33; // Forward declaration of class Mat33 - AT Feb 5, 2011

//Vec3 class - dedicated class for 3-dimensional vectors
class Vec3
{
	public:
		double vec[3];
		friend class Mat33;
		
	//Constructors and destructors
		Vec3();
		Vec3(const double, const double,const double);
		Vec3(const double*);
		Vec3(const Vec3 &);
		~Vec3() {};
	
	//Utility functions
		Vec3& operator = (const Vec3 &);
		void V3Zeros();
		void V3Swap(Vec3 &);
		
	//Comparison operators
		bool operator== (const Vec3 &); 
		bool operator!= (const Vec3 &); 

	//Level 1 BLAS
		Vec3 operator+ (const double);
		Vec3& operator+= (const double);
		Vec3 operator- (const double); 
		Vec3& operator-= (const double);
		Vec3 operator* (const double); 
		Vec3& operator*= (const double);
		Vec3 operator/ (const double); 
		Vec3& operator/= (const double);
		
		Vec3 operator+ (const Vec3 &); 
		Vec3& operator+= (const Vec3 &);
		Vec3 operator- (const Vec3 &); 
		Vec3& operator-= (const Vec3 &);
		double operator* (const Vec3 &) const;
	
	//Level 2 BLAS
		Vec3 operator* (const Mat33 &); 
		Vec3& operator*= (const Mat33 &);
		
	//Special functions
		Vec3 V3Cross(const Vec3 &);
		Mat33 V3TensProd(const Vec3 &);
		double V3Norm() const; 
		double V3Sum() const; 
		double V3Prod() const; 
		std::string V3Str() const; 
};

//Mat33 class - for when you need to do transformations on 3-vectors…
class Mat33
{
	public:
		double mat[3][3];
		friend class Vec3;
		
		//Constructors and destructors
		Mat33();
		Mat33(const Mat33 &);
		Mat33(const Vec3 &, const Vec3 &);
		Mat33(const double);
		~Mat33() {};
		
		//Utility functions
		Mat33& operator= (const Mat33 &);
		void M3Eye();
		void M3Zeros();
		
		//Comparison operators
		bool operator== (const Mat33 &);
		bool operator!= (const Mat33 &);
		 
		//Matrix-scalar operations
		Mat33 operator+ (const double);
		Mat33& operator+= (const double);
		Mat33 operator- (const double);
		Mat33& operator-= (const double);
		Mat33 operator* (const double);
		Mat33& operator*= (const double);
		Mat33 operator/ (const double);
		Mat33& operator/= (const double);
		
		//Level 2 BLAS
		Vec3 operator* (const Vec3 &);
		//No *= operator due to dataype mismatch.
		
		//Level 3 BLAS
		Mat33 operator+ (const Mat33 &);
		Mat33& operator+= (const Mat33 &);
		Mat33 operator- (const Mat33 &);
		Mat33& operator -= (const Mat33 &);
		Mat33 operator* (const Mat33 &);
		Mat33& operator*= (const Mat33 &);
		//These do not compile properly. Ask someone who knows C++ better than I do. - LEJ 01/07/10
		/*Mat33 operator/ (const Mat33 &);
		Mat33& operator/= (const Mat33 &)*/;
		
		//Special functions
		Mat33 M3Transpose();
		Mat33 M3RowSwap(const int, const int);
		Mat33 M3Inv(bool);
		double M3Trace();
		double M3Det();
		Vec3 M3Diag();
		void M3GEPP(Vec3 &x);
		void M3BackSub(Vec3 &x);
		Vec3 M3LinSolve(Vec3 &u);
		std::string M3Str();
		std::string M3RowStr(const int);
};


/*	Function implementations are below. Math functions and constructors are inlined for speed.
 *	Non-math functions are kept in .cpp file. The vast majority of the class is contained in
 *	This file, organized in the same order that the prototypes are listed */


/*	***Vec3***	*/

//Constructors and Destructors

//Default constructor (zeros)
inline Vec3::Vec3()
{
	vec[0] = 0.0; vec[1] = 0.0; vec[2] = 0.0;
}

//Construct from doubles
inline Vec3::Vec3(const double a, const double b, const double c)
{
	vec[0] = a;
	vec[1] = b;
	vec[2] = c;
}

//Construct from an array of 3 doubles
inline Vec3::Vec3(const double* A)
{
	vec[0] = A[0];
	vec[1] = A[1];
	vec[2] = A[2];
}

//Copy constructor
inline Vec3::Vec3(const Vec3 &u)
{
	vec[0] = u.vec[0];
	vec[1] = u.vec[1];
	vec[2] = u.vec[2];
}


//Non-BLAS utility functions

//Zero vector v = [0 0 0]
inline void Vec3::V3Zeros()
{
	vec[0] = 0.0;
	vec[1] = 0.0;
	vec[2] = 0.0;
}

//Assignment operator
inline Vec3& Vec3::operator=(const Vec3 &u)
{
	vec[0] = u.vec[0];
	vec[1] = u.vec[1];
	vec[2] = u.vec[2];
	return *this;
}

//Swap values in two vectors u <-> v
inline void Vec3::V3Swap(Vec3 &u)
{
	double temp;
	for (int i = 0; i < 3; i ++)
	{
		temp = vec[i];
		vec[i] = u.vec[i];
		u.vec[i] = temp;
	}
}


//Comparison operators
inline bool Vec3::operator==(const Vec3 &u)
{
	if (u.vec[0] != vec[0]) return false;
	if (u.vec[1] != vec[1]) return false;
	if (u.vec[2] != vec[2]) return false;
	else return true;
}

inline bool Vec3::operator!=(const Vec3 &u)
{
	if (u.vec[0] != vec[0]) return true;
	if (u.vec[1] != vec[1]) return true;
	if (u.vec[2] != vec[2]) return true;
	else return false;
}


//Level 1 BLAS

//Vector scalar addition v = u + a
inline Vec3 Vec3::operator+ (const double a)
{
	Vec3 v(a+vec[0], a+vec[1], a+vec[2]);
	return v;
}

//Vector scalar addition u = u + a
inline Vec3& Vec3::operator += (const double a)
{
	vec[0] +=a;
	vec[1] +=a;
	vec[2] +=a;
	return *this;
}

//Vector scalar subtraction v = u - a
inline Vec3 Vec3::operator- (const double a)
{
	Vec3 v(vec[0]-a, vec[1]-a, vec[2]-a);
	return v;
}

//Vector scalar subtraction u = u - a
inline Vec3& Vec3::operator -= (const double a)
{
	vec[0] -=a;
	vec[1] -=a;
	vec[2] -=a;
	return *this;
}

//Vector scalar multiplication v = u*a
inline Vec3 Vec3::operator* (const double a)
{
	Vec3 v(a*vec[0], a*vec[1], a*vec[2]);
	return v;
}

//Vector scalar multiplication u = u*a
inline Vec3& Vec3::operator*= (const double a)
{
	vec[0] *=a;
	vec[1] *=a;
	vec[2] *=a;
	return *this;
}

//Vector scalar division v = u/a
inline Vec3 Vec3::operator/ (const double a)
{
	double b = 1.0/a;
	Vec3 v(vec[0]*b, vec[1]*b, vec[2]*b);
	return v;
}

//Vector scalar division u = u/a
inline Vec3& Vec3::operator/= (const double a)
{
	double b = 1.0/a;
	vec[0] *=b;
	vec[1] *=b;
	vec[2] *=b;
	return *this;
}

//Vector addition v = u + w
inline Vec3 Vec3::operator+ (const Vec3 &w)
{
	Vec3 v(vec[0]+w.vec[0], vec[1]+w.vec[1], vec[2] +w.vec[2]);
	return v;
}

//Vector addition u = u + w
inline Vec3& Vec3::operator+= (const Vec3 &w)
{
	vec[0] += w.vec[0];
	vec[1] += w.vec[1];
	vec[2] += w.vec[2];
	return *this;
}

//Vector subtraction v = u - w
inline Vec3 Vec3::operator- (const Vec3 &w)
{
	Vec3 v(vec[0]-w.vec[0], vec[1]-w.vec[1], vec[2] -w.vec[2]);
	return v;
}

//Vector subtraction u = u - w
inline Vec3& Vec3::operator-= (const Vec3 &w)
{
	vec[0] -= w.vec[0];
	vec[1] -= w.vec[1];
	vec[2] -= w.vec[2];
	return *this;
}

//Dot product a = v'*u
inline double Vec3::operator* (const Vec3 &B) const
{
	double a = vec[0]*B.vec[0] + vec[1]*B.vec[1] + vec[2]*B.vec[2];
	return a;
}


//Level 2 BLAS

//Vector-Matrix Multiply v = u'*A
inline Vec3 Vec3::operator* (const Mat33 &A)
{
	Vec3 v(vec[0]*A.mat[0][0]+vec[1]*A.mat[0][1]+vec[2]*A.mat[0][2],
	vec[0]*A.mat[1][0]+vec[1]*A.mat[1][1]+vec[2]*A.mat[1][2],
	vec[0]*A.mat[2][0]+vec[1]*A.mat[2][1]+vec[2]*A.mat[2][2]);
	return v;
}

//Vector-Matrix multiply u' = u'*A
inline Vec3& Vec3::operator*= (const Mat33 &A)
{
	Vec3 u(*this);
	vec[0] = u.vec[0]*A.mat[0][0]+u.vec[1]*A.mat[0][1]+u.vec[2]*A.mat[0][2];
	vec[1] = u.vec[0]*A.mat[1][0]+u.vec[1]*A.mat[1][1]+u.vec[2]*A.mat[1][2];
	vec[2] = u.vec[0]*A.mat[2][0]+u.vec[1]*A.mat[2][1]+u.vec[2]*A.mat[2][2];
	return *this;
}


//Special functions

//Vector norm |v|
inline double Vec3::V3Norm() const
{
	double a = sqrt(vec[0]*vec[0]+vec[1]*vec[1]+vec[2]*vec[2]);
	return a;
}

//Cross product v = u x w
inline Vec3 Vec3::V3Cross(const Vec3 &w)
{
	Vec3 v(vec[0] = vec[1]*w.vec[2] - vec[2]*w.vec[1],
		 vec[1] = vec[2]*w.vec[0] - vec[0]*w.vec[2],
		 vec[2] = vec[0]*w.vec[1] - vec[1]*w.vec[0]);
	return v;
}

//Tensor product A = u*w'
inline Mat33 Vec3::V3TensProd(const Vec3 &w)
{
	Mat33 A(0.0);
	
	for(int i = 0; i < 3; i++)
	{
		A.mat[i][0] += vec[i]*w.vec[0];
		A.mat[i][1] += vec[i]*w.vec[1];
		A.mat[i][2] += vec[i]*w.vec[2];
	}
	
	return A;
}

//Sum of vector elements
inline double Vec3::V3Sum() const
{
	double a = vec[0]+vec[1]+vec[2];
	return a;
}

//Product of vector elements
inline double Vec3::V3Prod() const
{
	double a = vec[0]*vec[1]*vec[2];
	return a;
}




/*	***Mat33***	*/

//Constructors and destructors

//Default constructor (identity matrix)
inline Mat33::Mat33(void)
{
	mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0;
	mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0;
	mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0;
}

//Copy constructor
inline Mat33::Mat33(const Mat33 &A)
{
	for (int i = 0; i < 3; i++)
	{
		mat[i][0] = A.mat[i][0];
		mat[i][1] = A.mat[i][1];
		mat[i][2] = A.mat[i][2];
	}
}

//Tensor constructor
inline Mat33::Mat33(const Vec3 &u, const Vec3 &w)
{
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] = u.vec[i]*w.vec[0];
		mat[i][1] = u.vec[i]*w.vec[1];
		mat[i][2] = u.vec[i]*w.vec[2];
	}	
}

//Constant constructor
inline Mat33::Mat33(const double a)
{
	for (int i = 0; i < 3; i++)
	{
		mat[i][0] = a;
		mat[i][1] = a;
		mat[i][2] = a;
	}
}


//Utility functions

//Assignment operator operator B = A
inline Mat33& Mat33::operator= (const Mat33 &A)
{
	for (int i = 0; i < 3; i++)
	{
		mat[i][0] = A.mat[i][0];
		mat[i][1] = A.mat[i][1];
		mat[i][2] = A.mat[i][2];
	}
	return *this;
}

//Identity matrix
inline void Mat33::M3Eye()
{
	mat[0][0] = 1.0; mat[0][1] = 0.0; mat[0][2] = 0.0;
	mat[1][0] = 0.0; mat[1][1] = 1.0; mat[1][2] = 0.0;
	mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 1.0;
}

//Zero matrix
inline void Mat33::M3Zeros()
{
	mat[0][0] = 0.0; mat[0][1] = 0.0; mat[0][2] = 0.0;
	mat[1][0] = 0.0; mat[1][1] = 0.0; mat[1][2] = 0.0;
	mat[2][0] = 0.0; mat[2][1] = 0.0; mat[2][2] = 0.0;
}


//Comparison operators
inline bool Mat33::operator== (const Mat33 &A)
{
	for(int i = 0; i < 3; i++)
	{
		if (A.mat[i][0] != mat[i][0]) return false;
		if (A.mat[i][1] != mat[i][1]) return false;
		if (A.mat[i][2] != mat[i][2]) return false;
	}
	return true;
}

inline bool Mat33::operator!= (const Mat33 &A)
{
	for(int i = 0; i < 3; i++)
	{
		if (A.mat[i][0] == mat[i][0]) return false;
		if (A.mat[i][1] == mat[i][1]) return false;
		if (A.mat[i][2] == mat[i][2]) return false;
	}
	return true;
}


//Matrix-scalar operations

//Matrix-scalar addition B = A + a
inline Mat33 Mat33::operator+ (const double a)
{
	Mat33 B(*this);
	for(int i = 0; i < 3; i++)
	{
	B.mat[i][0] += a;
	B.mat[i][1] += a;
	B.mat[i][2] += a;
	}
	return B;
}

//Matrix-scalar addition A = A + a
inline Mat33& Mat33::operator+= (const double a)
{
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] += a;
		mat[i][1] += a;
		mat[i][2] += a;
	}
	return *this;
}

//Matrix-scalar subtraction B = A - a
inline Mat33 Mat33::operator- (const double a)
{
	Mat33 B(*this);
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] -= a;
		B.mat[i][1] -= a;
		B.mat[i][2] -= a;
	}
	return B;
}

//Matrix-scalar subtraction A = A - a
inline Mat33& Mat33::operator-= (const double a)
{
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] -= a;
		mat[i][1] -= a;
		mat[i][2] -= a;
	}
	return *this;
}

//Matrix-scalar multiplication B = A*a
inline Mat33 Mat33::operator* (const double a)
{
	Mat33 B(*this);
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] *= a;
		B.mat[i][1] *= a;
		B.mat[i][2] *= a;
	}
	return B;
}

//Matrix-scalar multiplication A = A*a
inline Mat33& Mat33::operator*= (const double a)
{
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] *= a;
		mat[i][1] *= a;
		mat[i][2] *= a;
	}
	return *this;
}

//Matrix-scalar division B = A/a
inline Mat33 Mat33::operator/ (const double a)
{
	Mat33 B(*this);
	double inva = 1.0/a;
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] *= inva;
		B.mat[i][1] *= inva;
		B.mat[i][2] *= inva;
	}
	return B;
}

//Matrix-scalar division A = A/a
inline Mat33& Mat33::operator/= (const double a)
{
	double inva = 1.0/a;
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] *= inva;
		mat[i][1] *= inva;
		mat[i][2] *= inva;
	}
	return *this;
}


//Level 2 BLAS

//Matrix-vector multiplication v = A*u
inline Vec3 Mat33::operator* (const Vec3 &u)
{
	Vec3 v(mat[0][0]*u.vec[0]+mat[0][1]*u.vec[1]+mat[0][2]*u.vec[2],
		   mat[1][0]*u.vec[0]+mat[1][1]*u.vec[1]+mat[1][2]*u.vec[2],
		   mat[2][0]*u.vec[0]+mat[2][1]*u.vec[1]+mat[2][2]*u.vec[2]);
	return v;
}


//Level 3 BLAS

//Matrix addition B = A + C
inline Mat33 Mat33::operator+ (const Mat33 &C)
{
	Mat33 B(*this);
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] += C.mat[i][0];
		B.mat[i][1] += C.mat[i][1];
		B.mat[i][2] += C.mat[i][2];
	}
	return B;
}

//Matrix addition C = A + C
inline Mat33& Mat33::operator+= (const Mat33 &C)
{
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] += C.mat[i][0];
		mat[i][1] += C.mat[i][1];
		mat[i][2] += C.mat[i][2];
	}
	return *this;
}

//Matrix subtraction B = A - C
inline Mat33 Mat33::operator- (const Mat33 &C)
{
	Mat33 B(*this);
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] -= C.mat[i][0];
		B.mat[i][1] -= C.mat[i][1];
		B.mat[i][2] -= C.mat[i][2];
	}
	return B;
}

//Matrix subtraction A = A - C
inline Mat33& Mat33::operator -= (const Mat33 &C)
{
	for(int i = 0; i < 3; i++)
	{
		mat[i][0] -= C.mat[i][0];
		mat[i][1] -= C.mat[i][1];
		mat[i][2] -= C.mat[i][2];
	}
	return *this;
}

//Matrix multiplication B = A*C
inline Mat33 Mat33::operator* (const Mat33 &C)
{
	Mat33 B(0.0);
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] += mat[i][0]*C.mat[0][0] + mat[i][1]*C.mat[1][0] + mat[i][2]*C.mat[2][0];
		B.mat[i][1] += mat[i][0]*C.mat[0][1] + mat[i][1]*C.mat[1][1] + mat[i][2]*C.mat[2][1];
		B.mat[i][2] += mat[i][0]*C.mat[0][2] + mat[i][1]*C.mat[1][2] + mat[i][2]*C.mat[2][2];
	}
	return B;
}

//Matrix multiplication A = A*C
inline Mat33& Mat33::operator*= (const Mat33 &C)
{
	Mat33 A(*this);
	for(int i = 0; i < 3; i++)
	{
		for(int j = 0; j < 3; j++)
		{
			mat[i][j] = A.mat[i][0]*C.mat[0][j] + A.mat[i][1]*C.mat[1][j] + A.mat[i][2]*C.mat[2][j];
		}
	}
	return *this;
}

//Matrix right division B = A/C
//Does not compile properly. LEJ 01/07/2010
/*inline Mat33 Mat33::operator/ (const Mat33 &C)
{
	Mat33 B(0.0);
	Mat33 D = C.M3Inv(NULL);
	for(int i = 0; i < 3; i++)
	{
		B.mat[i][0] += mat[i][0]*D.mat[0][0] + mat[i][1]*D.mat[1][0] + mat[i][2]*D.mat[2][0];
		B.mat[i][1] += mat[i][0]*D.mat[0][1] + mat[i][1]*D.mat[1][1] + mat[i][2]*D.mat[2][1];
		B.mat[i][2] += mat[i][0]*D.mat[0][2] + mat[i][1]*D.mat[1][2] + mat[i][2]*D.mat[2][2];
	}
	return B;
}

//Matrix right division A = A/C
inline Mat33& Mat33::operator/= (const Mat33 &C)
{
	Mat33 A(*this);
	Mat33 D = C.M3Inv(NULL);
	for(int i = 0; i < 3; i++)
	{
		for(int j = 0; j < 3; j++)
		{
			mat[i][j] = A.mat[i][0]*C.mat[0][j] + A.mat[i][1]*C.mat[1][j] + A.mat[i][2]*C.mat[2][j];
		}
	}
	return *this;
}*/
								

//Special functions

//Matrix determinant d = Det(A)
inline double Mat33::M3Det()
{
	double det = mat[0][0]*(mat[1][1]*mat[2][2]-mat[1][2]*mat[2][1]);
	det += mat[1][0]*(mat[2][1]*mat[0][2]-mat[0][1]*mat[2][2]);
	det += mat[2][0]*(mat[0][1]*mat[1][2]-mat[0][2]*mat[1][1]);
	return det;
}

//Matrix transpose C = A'
inline Mat33 Mat33::M3Transpose()
{
	Mat33 C;
	for(int i = 0; i < 3; i++)
	{
		C.mat[i][0] = mat[0][i];
		C.mat[i][1] = mat[1][i];
		C.mat[i][2] = mat[2][i];
	}
	return C;
}

//Matrix row swap
inline Mat33 Mat33::M3RowSwap(const int a, const int b)
{
	Mat33 A;
	double temp;
	for (int j = 0; j < 3; j++)
	{
		temp = mat[b][j];
		A.mat[b][j] = mat[a][j];
		A.mat[a][j] = temp;
	}
	return A;
}

//Extract diagonal elements of matrix v = Aii -- goes in Mat33 class
inline Vec3 Mat33::M3Diag()
 {
 Vec3 v(mat[0][0],mat[1][1],mat[2][2]);
 return v;
 }

//Trace of a matrix a = Tr(A)
inline double Mat33::M3Trace()
{
	double a = mat[0][0]+mat[1][1]+mat[2][2];
	return a;
}

//Back-substitution on 3x3 triangular system
inline void Mat33::M3BackSub(Vec3 &x)
{
	x.vec[2] = x.vec[2]/mat[2][2];
	x.vec[1] = (x.vec[1] - mat[1][2]*x.vec[2])/mat[1][1];
	x.vec[0] = (x.vec[0] - mat[0][1]*x.vec[1] - mat[0][2]*x.vec[2])/mat[0][0];
}

	//Cramer's rule inversion for 3x3 matrix

inline Mat33 Mat33::M3Inv(bool sing)
{
	Mat33 A(*this);
	
	double det = A.M3Det();
	
	double ad = fabs(det);
	if (ad < 1.0e-10)
	{
		printf("WARNING: Matrix is close to singular. det = %f\n", det);
		sing = 1;
	}
	
	double invdet = 1.0/det;
	
	A.mat[0][0] = (mat[1][1]*mat[2][2]-mat[2][1]*mat[1][2])*invdet;
	A.mat[0][1] = (mat[2][1]*mat[0][2]-mat[0][1]*mat[2][2])*invdet;
	A.mat[0][2] = (mat[0][1]*mat[1][2]-mat[1][1]*mat[0][2])*invdet;
	
	A.mat[1][0] = (mat[2][0]*mat[1][2]-mat[1][0]*mat[2][2])*invdet;
	A.mat[1][1] = (mat[0][0]*mat[2][2]-mat[2][0]*mat[0][2])*invdet;
	A.mat[1][2] = (mat[1][0]*mat[0][2]-mat[0][0]*mat[1][2])*invdet;
	
	A.mat[2][0] = (mat[1][0]*mat[2][1]-mat[1][1]*mat[2][0])*invdet;
	A.mat[2][1] = (mat[2][0]*mat[0][1]-mat[0][0]*mat[2][1])*invdet;
	A.mat[2][2] = (mat[0][0]*mat[1][1]-mat[1][0]*mat[0][1])*invdet;
	
	return A;
}

//Linear Solver for 3x3 systems using GEPP + back-substitution
inline Vec3 Mat33::M3LinSolve(Vec3 &u)
{
	Mat33 A(*this);
	Vec3 v(u);
	A.M3GEPP(v);
	A.M3BackSub(v);
	return v;	
}

