#include "evolve_system.h"

/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 */

//Core function for Metropolis MC calculation.
//Starts with an array of ellipsoids (BeOids) and evolves it, with the outer (step) loop being over MC cycles,
//with Noids moves per cycle, and the inner (ii) loop being individual steps within each cycle. Ellipsoids are
//moved randomly (not in order), with the ellipsoid selected for a step being given the index kk.
//Data is saved and macroscopic parameters updated every *cycle*
void evolve_system(Molecule** BeOids)
{
	//Initialize clock and output
	printf("\nSetup complete. Starting simulation...\n");
	double tstart = clock();
	output_header();
	
	Cnsts *cnsts = GetCnsts();
	
	//Allocate memory
	Vec3* dipoles = new Vec3[cnsts->n_oids];	//Stores rotated (lab frame) dipole moments of individual ellipsoids
	Vec3* cosmeans = new Vec3[cnsts->steps];	//Stores order parameters
	Vec3* Ms = new Vec3[cnsts->steps];	//Stores total dipole moment of system
	Vec3* totalV = new Vec3[cnsts->steps];	//Stores potential energy components for systems
	Vec3 Tmpmu, expecV, Vstep, Evec;
	
	double epsRF;	//Reaction field dielectric constant
	double oldmuE = 0.0, muE = 0.0;
	double oldmumu = 0.0 , Vmumu = 0.0;
	double oldVLJ  = 0.0, VLJ = 0.0;
	double tmovtemp, rmovtemp, tmovstep, rmovstep;
	double msmoved[2] = {0.0, 0.0};
	double OldV, NewV, deltaV, totVstep, nMstep;
	double moving_oids, rotating_oids;
	double iNkT; // 1/NkT
	
	long accept[3] = {0,0,0}; //Move acceptance counters

	
	//Define flags and constants;
	bool skipmumu = 1, recalc_muE = 0;
	long step = cnsts->steps;
	long eqsteps = cnsts->steps-cnsts->avg_steps; //Pre-equilibrium cycles before simulation starts recording data
	
	//generate file for g(r) if dynamic g(r) is enabled
	if (cnsts->dogr == 1) setup_gr();
	
	//generate extended means file if extended means are enabled
	if (cnsts->doxm == 1) setup_extended_means();
	
	//generate A and B means file if enabled
	if (cnsts->binaryavg == 1) setup_binary_means();
		
	//Calculate initial dipole configuration
	get_dipoles(BeOids, dipoles);
	
	//Allocate memory for choosing moves. nA and nB need to be explicity typecast to double from int.
	int kk;
	double dkk;
	double nA = (double)cnsts->num_A;
	double nB = (double)cnsts->num_B;
	
	//Determine number of oids that can move or rotate;
	moving_oids = nA - nA*cnsts->rot_notransA + nB - nB*cnsts->rot_notransB;
	rotating_oids = nA - nA*cnsts->stillmembersA + nB - nB*cnsts->stillmembersB;
	
	//Set initial dielectric conditions and calculate initial reaction field
	epsRF = cnsts->epsilon;
	cnsts->rfcorr = 2.0*(epsRF-cnsts->n2)/(qpwr(cnsts->rfcut,3)*(2.0*epsRF+cnsts->n2));
	
	//This is it! Start the main (steps) loop
	for (step=0;step<cnsts->steps;step++)
	{
		//Reset displacement counter to zero
		tmovstep = 0.0;
		rmovstep = 0.0;
		
		//Update dielectric if using dynamic reaction field, and calculate the reaction field coefficient
		//Note that this requires recalculating mu-mu interactions, causing a performance hit
		epsRF = update_eps(dipoles);
		if (cnsts->scrf == 1 && skipmumu == 0)
		{
			cnsts->rfcorr = 2.0*(epsRF-cnsts->n2)/(qpwr(cnsts->rfcut,3)*(2.0*epsRF+cnsts->n2));
			recalcmm(BeOids, dipoles, Vstep);
		}
		
		//check if E and T need to be updated, and update uE energy if E is updated
		recalc_muE = setEandT(Evec, step, epsRF);
		if (recalc_muE == 1 && skipmumu == 0)
		{
			Vstep.vec[0] = VmuEtotal(dipoles, Evec);
			recalc_muE = 0;
		}
		iNkT = 1.0/(cnsts->n_oids*cnsts->kT);
		
		//Calculate first step energy if this is the first step
		if (step == 0)
		{
			firststep(BeOids, dipoles, Evec, Vstep, 0);
			totVstep = Vstep.V3Sum();
			printf("Initial energy components (NkT):\nVmuE		Vmumu		VLJ		Total\n%f	%f	%f	%f\n", Vstep.vec[0]*iNkT, Vstep.vec[1]*iNkT, Vstep.vec[2]*iNkT, totVstep*iNkT);

			//Check to see if randomization is used
			if (cnsts->randsteps != 0)
			{
				skipmumu = 1;
				printf("Disabling electrostatics and randomizing system...\n\n");
				
				//Discard everything except the VdW term, so energies don't get
				//unphysically propagated during the randomization stage
				Vstep.vec[0] = 0.0;
				Vstep.vec[1] = 0.0;
			}
			else 
			{
				skipmumu = 0;
				printf("Randomization phase skipped.\n\n");
			}
			
			//Set up columns for remainder of log
			printf("STEP		U (NkT)			NORMALIZED ORDER\n");
		}
		
		//Recalculate total energy once mu-mu interactions are turned on and display energy/timing update
		else if (step == cnsts->randsteps)
		{
			skipmumu = 0;
			cnsts->Rand_Time=(clock()-tstart)/CLOCKS_PER_SEC;
			printf("\nInitial randomization complete. Randomization took %f seconds.\n",(clock()-tstart)/CLOCKS_PER_SEC);
			printf("Initial acceptance rate (%%): %f\n", (double)accept[0]/(double)(cnsts->randsteps*cnsts->n_oids)*100.0);
			printf("Initial RMS move sizes: %f (A)	%f (rad)\n", sqrt(msmoved[0]/cnsts->randsteps), sqrt(msmoved[1]/cnsts->randsteps));
			printf("\nEnabling electrostatics and recalculating total energy.\n");
			Vstep.vec[0] = VmuEtotal(dipoles, Evec);
			recalcmm(BeOids, dipoles, Vstep);
			totVstep = Vstep.V3Sum();
			printf("Energy components (NkT):\nVmuE		Vmumu		VLJ		Total\n%f	%f	%f	%f\n\n", Vstep.vec[0]*iNkT, Vstep.vec[1]*iNkT, Vstep.vec[2]*iNkT, totVstep*iNkT);
			printf("STEP		U (NkT)			NORMALIZED ORDER\n");
		}
		
		//Inner (oids) loop -> do n_oids moves every MC cycle, picking oids randomly
		for (int ii=0;ii<cnsts->n_oids;ii++)
		{
			
			//Randomly select molecule with index kk. For completely general code, this should be a global function that handles any number of types.
	        if((cnsts->stillmembersA==0)&&(cnsts->stillmembersB==0)) dkk=(nA+nB)*ran2();
	        else if(cnsts->stillmembersA==1) dkk=nA+nB*ran2();
	        else if(cnsts->stillmembersB==1) dkk=nA*ran2();
            kk=(int)dkk;
			
			//create temporary copy of chosen molecule.
			Molecule* Tmpoid = new Molecule(*BeOids[kk]);
			Tmpmu = dipoles[kk];
			
			//Calculate mu-mu and LJ energy
			OneStep(BeOids, dipoles, kk, oldmumu, oldVLJ, skipmumu);
			
			//Calculate current muE energy and total the energies
			if (step >= cnsts->randsteps) oldmuE = VmuE(dipoles[kk], Evec);
			OldV = oldmuE + oldmumu + oldVLJ;

			//Move the chosen molecule. Incorporate conditional into class later.
	        if (BeOids[kk]->rot_notrans==0)
			{
				if (cnsts->latticetype == 1) tmovtemp = BeOids[kk]->Translate();
				else if (cnsts->latticetype == 3) tmovtemp = BeOids[kk]->Translate_Sphere();
				else if (cnsts->latticetype == 4) tmovtemp = BeOids[kk]->Translate_Cylinder();
			    else BeOids[kk]->Translate();
		    } 
			rmovtemp = BeOids[kk]->Rotate();
			
			//Update the dipoles array. The saved dipole will be used to revert if the change is rejected.
			update_dipoles(BeOids[kk], dipoles[kk]);

			//Recalculate energy
			OneStep(BeOids, dipoles, kk, Vmumu, VLJ, skipmumu);
			
			//Calculate new muE energy and total energies
			if (step >= cnsts->randsteps) muE = VmuE(dipoles[kk], Evec);
			NewV = muE + Vmumu + VLJ;

	        //Metropolis Conditional:
            deltaV = NewV - OldV;
            double rand = ran2();

	        if(exp(-1.0*deltaV*cnsts->beta) > rand)
			{
				
				//Update energy
		        BeOids[kk]->LastV=NewV;
				
				//Record squared displacement
				tmovstep += tmovtemp;
				rmovstep += rmovtemp;
				
				//acceptance tracking [vdw preeq eq]
				if (step>=eqsteps) accept[2]+=1;
				else if (step>=cnsts->randsteps) accept[1]+=1;
				else accept[0]+=1;
	        }
			
			//Revert energies and positions if move is rejected
	        else
			{
			    BeOids[kk]->LastV=OldV;
				muE = oldmuE;
				Vmumu = oldmumu;
				VLJ = oldVLJ;
				
				dipoles[kk] = Tmpmu;
				BeOids[kk]->center = Tmpoid->center;
				BeOids[kk]->rot = Tmpoid->rot;
		    }
			
			//Record instantaneous value of total energy
			Vstep.vec[0] += (muE-oldmuE);
			Vstep.vec[1] += (Vmumu - oldmumu);
			Vstep.vec[2] += (VLJ - oldVLJ);
			
			//clean up
			delete Tmpoid;
			
	    }//end oids loop.
		
		
		//Output section. Saves data for processing and dumps messages to standard out.
		
		//Calculate cos^n, M, V
		get_means(BeOids, dipoles, cosmeans[step], Ms[step]);
		totalV[step] = Vstep;
		
		//Calculate mean squared displacement for step
		msmoved[0] += tmovstep/moving_oids;
		msmoved[1] += rmovstep/rotating_oids;

		//Output progress indicator to user, including total energy trace.
		long splus1 = step+1;
		if (splus1 % 100 == 0)
		{
			nMstep = sqrt(Ms[step]*Ms[step])/((cnsts->num_A*cnsts->muzA+cnsts->num_B*cnsts->muzB)*cnsts->mucorr);
			totVstep = Vstep.V3Sum()*iNkT;
			printf("\n%li		%f			%f\n", splus1, totVstep, nMstep);
			
			if (cnsts->anneal == 1 && step < cnsts->randsteps*2)
			{
				printf("			T = %f K\n",cnsts->kT/cnsts->kB);
			}
			if (cnsts->scrf == 1 && step > cnsts->randsteps)
			{
				printf("			Epsilon = %f\n", epsRF);
			}
			if (cnsts->realfield == 1 && step > cnsts->randsteps && Evec.vec[2] != 0)
			{
				printf("			E0z = %f V/um\n", Evec.vec[2]/cnsts->Escale);
			}
		}
		
		//Correlation function calculation ( g(r), etc)
		if (cnsts->dogr == 1 && splus1 > eqsteps && splus1 % cnsts->grfreq == 0)
		{
			printf("			Calculating correlation functions\n");
			gofr(BeOids);
		}
		
		//Extended means calculation (angles with respect to x, y, and z axis instead of just z)
		if (cnsts->doxm == 1 && splus1 > eqsteps && splus1 % cnsts->xmfreq == 0)
		{
			printf("			Calculating additional order parameters\n");
			get_extended_means(BeOids);
		}
		
		//Separate extended means calculation for types A and B
		if (cnsts->binaryavg == 1 && splus1 > eqsteps && splus1 % cnsts->xmfreq == 0)
		{
			printf("			Calculating A and B order parameters\n");
			get_binary_means(BeOids);
		}
		
		//Save intermediate coordinates. Note that this can require large amounts of disk space
		if (cnsts->chkcoord == 1 && splus1 > eqsteps && splus1 < cnsts->steps && splus1 % cnsts->grfreq == 0)
		{
			printf("			Recording coordinates\n");
			get_positions(BeOids, splus1, 0);
		}
		
		//Premature termination code for debugging purposes. Do NOT use for normal operation.
		/*if (step == 1)
		{
			printf("Terminating run at step %li and recording positions.\n", step);
			get_positions(BeOids, 0, 2);
			output_data(cosmeans, Ms, totalV, tstart, accept, msmoved, splus1);
			exit(12);
		}*/
		
	}//end steps loop
	
	//Get final positions and orientations of ellipsoids, then output data to simulation log
	get_positions(BeOids, cnsts->steps, 1);
	output_data(cosmeans, Ms, totalV, tstart, accept, msmoved, cnsts->steps);
	
	//clean up
	delete[] dipoles;
	delete[] cosmeans;
	delete[] Ms;
	delete[] totalV;
}


