/*
* Robinson Group Classical Monte Carlo Molecular Simulator (legacy version)
* Created by Lewis E. Johnson and Robin Barnes, Robinson Group, Department of Chemistry, University of Washington
* http://courses.washington.edu/bhrchem/
*
* Current build 05/16/2012; last major build 02/17/2010
* Released under the University of Illinois/NCSA Open Source License (http://otm.illinois.edu/uiuc_openSource)
*
* This code was used for simulations in Chapters 2, 3, and 4 of Lewis E. Johnson's dissertation, and has been kept as-used except
* for minor bugfixes and cosmetic changes to the console output. It has since been surpassed by much more powerful code
* developed by Andreas Tillack and Lewis Johnson, which is not yet available for public release.
*
* This software  can be compiled under Linux or OSX without modification. It currently only requires the standard C++ library,
* and has been tested with gcc 2.96, 3.43, 4.01, 4.12, 4.2, as well as newer versions of gcc, LLVM-gcc 4.2, LLVM-clang 1.6, icc 9,
* icc 10.1, and icc 11.1. Optimization at the -O2, -O3, or -fast level is recommended.
*
* Settings for a simulation are currently declared in the main file; recompiling is required between simulations.
* Linear algebra methods are declared in VecMat.h, stats methods are declared in utility_functions.h
* Parameters that do not need to be changed by the user, and most of the error-handling code are in Setup.cpp
* Output is generated as text files. Some are easier for Matlab to read than others.
* Simulation parameters are printed at beginning of run for ease of monitoring a simulation.
* Premature exit codes are (1) - initial parameters error, (2) - singular matrix in Touch, (3) Fully overlapped particle
* (invalid distance), (4) - coordinate file loading error, and (12) - User initiated exit for debugging purposes.
* Note that error type (3) is often the result of a corrupted build; try cleaning and rebuilding before digging into the code.
*
* This software is based on prior Monte Carlo simulation code by Bruce Robinson, et al. Development was funded by the NSF
* STC-MDITR (DMR 0120967). Further thanks are due to Andreas Tillack for assistance with debugging.
*
*/

#include "setup.h"
#include "Latticeoids3d.h"
#include "evolve_system.h"

int main()
{
	printf("\nRobinson Group Classical Monte Carlo Molecular Simulator\nBuild %s (development)", BUILD);
	Cnsts* cnsts = GetCnsts();
	
	//Specify number of runs and simulation name
	int runs = 1;	//number of runs in of batch
	char FileOut[] = "testOLD4_";	//stem of name for output files
	
	//Use test mode only for validating changes to code--it locks the RNG seeds and limits
	//batches to no more than two runs
	bool test = 1;
	long testseeds[2] = {-7, -1};
	
	//Input file info (used for loader = 3 ONLY). Put just the stem of the file(s) here.
	//Batch numbering will be handled automatically.
	char FileIn[] = "";	//stem of name for input files
	cnsts->chkpt = "final"; //Put "final" here for a chkfinalcoords file, or the step number (in quotes) if using intermediate checkpoint.
	
	//Initalize simulation control flags
	cnsts->dogr = 1;	//dynamically calculates correlation functions if 1
	cnsts->doxm = 1;	//calculates full <cos> vectors (instead of just <cos_z>) and <cos^2> tensors if 1
	cnsts->binaryavg = 1;	//Calculate separate <cos^n> for A and B if 1
	cnsts->chkcoord = 0;	//Records intermediate coordinates every grfreq steps (DIAGNOSTIC)
	cnsts->latticetype = 1;	//Specifies initial simulation geometry. 1 and 2 are cubic, 3 is spherical, 4 is cylindrical
	cnsts->PBCs[0] = 1; cnsts->PBCs[1] = 1; cnsts->PBCs[2] = 1; //Periodic boundary conditions along each axis
	cnsts->cut = 1;	//switches short-range cutoffs on and off for two-body interactions
	//cnsts->cut_type = 1; //Cutoff type. 1 is spherical, 2 is rectangular, 3 is cylindrical
	cnsts->scrf = 1; //Dynamicaly updating reaction field (SCRF).
	cnsts->vdwtype = 4; //van der Waals type. 1 is simple, 2 is extended, 3 is soft-sphere, 4 is Touch, 5 is soft-ellipsoid
	cnsts->loader = 2; //Loader used for generating initial lattice. 1 is sequential, 2 is random, and 3 loads from a file.
	cnsts->offctrmu = 0; //Allow off-center dipoles. May run slightly slower.
	cnsts->realfield = 0; //Switch between external field being defined in a vacuum (E0) or as the real field in the media (E)
	cnsts->anneal = 1; //Use simulated annealing for melting initial lattice
	cnsts->polarize = 0;	//Use Onsager anisotropic polarizability (simple Lorentz correction on dipole moment). No effect if n2 = 1
	
	//Initialize simulation parameters
    cnsts->steps = 12000;	//total number of MC cycles
	cnsts->randsteps = 1000; //cycles used to break up lattice (VdW only), must be smaller than steps.
    cnsts->avg_steps = 6000;	//steps used for averaging
	cnsts->nndist = 4.43;	//initial nearest-neighbor distance (Angstroms)
	cnsts->n_oids = 216;	//particles in system
	cnsts->Efield = 0.0; //Efield in V/um
	cnsts->T = 293.0; //Temp in K
	cnsts->epsilon = 1.0; //External dielectric constant for reaction field. No reaction field if this is 1.
	cnsts->n2 = 1.80; //Square of refractive index of system, used for polarizable media.
	
	//Initialize molecule parameters
	cnsts->num_A = 64;	//# of A-type molecules
	cnsts->r0A[0] = 2.04; cnsts->r0A[1] = 2.04; cnsts->r0A[2] = 3.04; //Dimensions for A-type molecule (A)
	cnsts->muzA = 4.09;		//Dipole moment of A-type molecules (Debye)
	cnsts->vdwA = 0.03;		//VdW parameter for A-type molecules (pico-ergs)
	cnsts->rot_notransA = 0;	//If 1, A-type molecules are on-lattice but can rotate
	cnsts->stillmembersA = 0;	//If 1, A-type molecules are stationary
	cnsts->averageA = 1;	//If 1, A-type molecules will be included in <cos^n(theta)>
	cnsts->num_B = 152;	//All of these are the same, but for B-oids
	cnsts->r0B[0] = 1.85; cnsts->r0B[1] = 1.85; cnsts->r0B[2] = 3.70;
	cnsts->muzB = 2.04;
	cnsts->vdwB = 0.02;
	cnsts->averageB = 1;
	cnsts->rot_notransB = 0;
	cnsts->stillmembersB = 0;
	
	//Initialize box parameters
    cnsts->lin_dim[0]= 6; cnsts->lin_dim[1] = 6; cnsts->lin_dim[2] = 6; //Number of molecules per each side of lattice
	cnsts->boxlength[0] =cnsts->nndist*cnsts->lin_dim[0]; //Box lengths (autoset)
	cnsts->boxlength[1] =cnsts->nndist*cnsts->lin_dim[1];
	if (cnsts->latticetype == 4) cnsts->boxlength[2] = 4.43*cnsts->lin_dim[2]; //Special z-axis for cylinders (manual set)
	else cnsts->boxlength[2] =cnsts->nndist*cnsts->lin_dim[2];
	
	//Step size parameters. Decrease step size if acceptance too low, increase if too high. Keep the size of your
	//particles in mind when setting initial step sizes.
	cnsts->maxtrans = 0.15;	//maximum translational move size (±) in Angstroms
	cnsts->maxrot = 0.13*pi;	//maximum translational move size (±) in radians
	
	
	/*
	********************************************************************************
	***Anything below here should only be modified if you know what you are doing***
	********************************************************************************
	*/
		
	//Box radius, used for spherical and cylindrical geometries only
	cnsts->spherecylr = 21.99; //radius (Angstroms) for spherical and cylindrical simulation cells (latticetype = 3 or 4)
	
	//Cutoff distances (Angstroms) for simulations that use a truncated potential
	//cnsts->cut_dist[0] = cnsts->boxlength[0]/2.0; //cutoff distance for x or r
	//cnsts->cut_dist[0] = 3*cnsts->nndist;
	//cnsts->cut_dist[1] = cnsts->boxlength[1]/2.0; //cutoff distance for y (used for rectangle only)
	//cnsts->cut_dist[2] = cnsts->boxlength[2]/2.0; //cutoff distance for z (used for rectangle or cylinder only)
	cnsts->ljcut = 4.0*cnsts->nndist;	//Default Lennard-Jones cutoff distance is four neighbor shells
	cnsts->escut = cnsts->boxlength[0]/2.0;	//Default electrostatic cutoff distance is half the box length (max possible)
	
	//Additional LJ parameters (exponents and implicit solvent model)
	cnsts->LJexp[0] = 6; cnsts-> LJexp[1] = 12; //Lennard-Jones exponents (int datatype!)
	cnsts->Solvent[0] = -1.01; cnsts-> Solvent[1] = 0.1; cnsts-> Solvent[2] =12.43; //Parameters for Gaussian attenuating function. Ignored if using simple LJ function!
	//0 is -2.355/FWHM, 1 is the correction strength, 2 is rmin (sigma * 2^1/6)
	
	//Dipole offsets for off-center dipoles (x,y, and z)
	cnsts->m0A[0] = 0.00; cnsts->m0A[1] = 0.00; cnsts->m0A[2] = 0.00;
	cnsts->m0B[0] = 0.00; cnsts->m0B[1] = 0.00; cnsts->m0B[2] = 0.00;
	
	//Scaling factors for mu_mu and mu_E interactions have been deprecated, and are now incorporated into polarizabilty model.
	
	//Parameters for extended means and correlation functions
	cnsts->xmfreq = 100;	//Calculate extended means every this many cycles
	cnsts->grfreq = 100;	//Calculate g(r) every this many cycles
	cnsts->grinc = 200;	//bins used in calculating g(r)
	cnsts->rmax = cnsts->boxlength[0]/2;	//cutoff radius for g(r)
	cnsts->refdr = cnsts->rmax/cnsts->grinc;	//radial increment for g(r)
	
	if (cnsts->dogr == 1 && cnsts->latticetype > 3)
	{
		printf("g(r) is currently untested for latticetype > 3\n");
		exit(1);
	}

	//Set angle of fixed/still molecules -- ex: molecules sitting on surface.
	cnsts->stillrot[0][0] = 1.; cnsts->stillrot[0][1] = 0.; cnsts->stillrot[0][2] = 0.;
	cnsts->stillrot[1][0] = 0.; cnsts->stillrot[1][1] = 1.; cnsts->stillrot[1][2] = 0.;
	cnsts->stillrot[2][0] = 0.; cnsts->stillrot[2][1] = 0.; cnsts->stillrot[2][2] = 1.;
	cnsts->stilleuler [0] = 0.; cnsts->stilleuler [1] = 0.; cnsts->stilleuler [2] = 0.;

	//boundary parameters for LDBR sims. Not currently used for anything.
	cnsts->Vvdwboundary = 0.5;
	cnsts->uBoundary = 12;
	cnsts->rboundary = 5.55;
	cnsts->uSigma = 14.1;
	
	//Parameters that need type conversion and/or loading into Cnsts. Don't touch these.
	cnsts->fileout = FileOut;
	cnsts->filein = FileIn;
	cnsts->test = test;
	cnsts->runs = runs;
	
	//Run setup and validation functions for simulation parameters. If everything passes,
	//then we'll start loading molecules.
	printf("\nSetting up the simulation...\n");
	phys_cnsts();
	setup_vdw();
	setup_electrostatics();
	check_box();
	check_infile();
	final_validation();
	printf("Setup complete. Loading parameters...\n");

	//Molecule parameters.
	//double a1[] = cnsts->r0A; //saxes for molecule type 0.
	double d1[] = {0.,0.,cnsts->muzA*cnsts->mucorr,cnsts->m0A[0],cnsts->m0A[1],cnsts->m0A[2]}; //dipole for molecule type 0.
	double v1 = cnsts->vdwA; // Vvdw for molecule type1
	
	Ellipse E1(cnsts->r0A, d1, v1); //make an instance of ellipse type 0.
	Molecule* s = new Molecule(E1,cnsts->stillmembersA,cnsts->rot_notransA, 0, cnsts->averageA);


	//double a2[3] = cnsts->r0B; //saxes for molecule type 1.
	double d2[] = {0.,0.,cnsts->muzB*cnsts->mucorr,cnsts->m0B[0], cnsts->m0B[1], cnsts->m0B[2]}; //dipole for molecule type 1.
	double v2 = cnsts->vdwB; //Vvdw for molecule type 1.
	
	Ellipse E2(cnsts->r0B, d2, v2); //make an instance of ellipse type 1.
    Molecule* t = new Molecule(E2,cnsts->stillmembersB,cnsts->rot_notransB, 1, cnsts->averageB);
	
	//Done with loading parameters. Prepare to run!
	Molecule** BeOids = new Molecule*[cnsts->n_oids];
	

	//Get current time and initialize memory for RNG seed
	printf("Initializing RNG seed...\n");
	long tempseed;
	time_t rawtime;
	struct tm * batchstarttime;
	time (&rawtime);
	batchstarttime = localtime (&rawtime);
	
	//Open batch summary file
	printf("Opening batch summary file...\n");
	open_summary();
	printf("Done. Batch started on %s\n",asctime(batchstarttime));
	
	//This is it! Execute the batch
	for (int sim=0; sim < runs; sim++)
	{
		//Reset system
		cnsts->trackchange = 0; //reset E and T to initial.
		if (test == 0) tempseed = -1*time(NULL); //get RNG seed from time. This won't repeat until 2038 :D
		else
		{
			//Locked seed for test mode
			printf("DIAGNOSTIC MODE ON\n");
			tempseed = testseeds[sim];
		}
		cnsts->rngseed = tempseed; //record the seed
		cnsts->idum = &tempseed; //idum changes each time ran2() is called
		cnsts->simnum++;
		
		//Run simulation
		LatticeOids3d* lattice = new LatticeOids3d(BeOids,*s,*t);   // This constructor loads molecules into BeOids.
		check_oids(BeOids);
		evolve_system(BeOids);
		delete lattice;
	}
	
	//Get ending time
	struct tm * batchendtime;
	time (&rawtime);
	batchendtime = localtime (&rawtime);
	
	//Clean up
	delete s;
	delete t;
	string summfn = cnsts->fileout + "summary.dat";
	FILE* summ = fopen(summfn.c_str(), "a");
	fprintf(summ, "\n#END\n");
	fclose(summ);
	
	//Normal termination
	printf("\nBatch finished on %s\nDone.\n", asctime(batchendtime));
	return 0;
}


