/*
 *
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *
 *  get_means.cpp
 *  Created by ljohnson on 12/30/08.
 */

#include "get_means.h"

//This set of functions is used to obtain mean values of order parameters and values of the dielectric constant.
//While most functions are called during a run, some may only be called from output_data.

//Calculation of total moments and first three moments of cos(theta) wrt the z axis. LEJ 12/30/2008
//Inputs:
//	BeOids - the array of all molecules in the system.
//	mus - the array of all dipole moment vectors in the system
//	temp_means - storage array for order parameters (holds till written to file)
//	M - storage array for total dipole moment (holds till written to file)
void get_means(Molecule** BeOids, Vec3* mus, Vec3 &temp_means, Vec3 &M)
{
	Cnsts *cnsts = GetCnsts();
	
	//Allocate memory
	double zproj, zp2;
	double avgmembers = (double)(cnsts->n_avg);
	Vec3 tmpmeans;
	M.V3Zeros();
	
	//Obtain cos(theta) and m values from the rotation matrices
	for (int i=0;i<cnsts->n_oids;i++)
	{
		//Expectation values can be set to exclude solvent molecules
		if (BeOids[i]->averageme == 1)
		{
			//get the moments of cos relative to the Z axis [0 0 1]. Use of any
			//other axis requires the extended means function.
			zproj = BeOids[i]->rot.mat[2][2];
			zp2 = zproj*zproj;
			tmpmeans.vec[0] +=zproj;
			tmpmeans.vec[1] +=zp2;
			tmpmeans.vec[2] +=zproj*zp2;
		}
		
		//Add up m values to obtain M
		M+=mus[i];
	}
	
	//Calculate the averages
	temp_means = tmpmeans/avgmembers;	
}

//SCRF Update; called if cnsts->scrf = 1
//Updates reaction field dielectric constant. Convergence may be difficult for poorly-behaved systems like strongly polar Stockmayer fluids.
//Inputs: mus - array of all lab-frame dipole moments
//Return value is the current reaction field dielectric constant
double update_eps(Vec3* mus)
{
	Cnsts *cnsts = GetCnsts();
	double epsilon, q;
	double n2 = cnsts->n2;
	double n4 = n2*n2;
	
	//Variance calculation if no field
	if (cnsts->Efield == 0)
	{
		Vec3 Mtemp;
	
		//Obtain current value of M^2
		for (int i = 0; i < cnsts->n_oids; i++)
		{
			Mtemp += mus[i];
		}
		
		double M2 = Mtemp*Mtemp;
	
		//Calculate q = yg
		q = 4*pi*M2/(9*cnsts->V*cnsts->kT);
		epsilon = 0.25*n2+2.25*q+0.75*sqrt(n4+2.0*n2*q+9.0*q*q);

	}
	
	//Calculation if E0z is present
	else
	{
		//Only need to sum z components of total dipole
		double Mztemp = 0.0;
		for (int j = 0; j < cnsts->n_oids; j++)
		{
			Mztemp += mus[j].vec[2];
		}
		
		//Calculate epsilon
		if (cnsts->realfield == 1)
		{
			epsilon = 4.0*pi*Mztemp/(cnsts->Efield*cnsts->Escale*cnsts->V) + cnsts->n2;
		}
		else
		{
			q = (4.0*pi*Mztemp)/(3.0*cnsts->V*cnsts->Efield*cnsts->Escale);
			epsilon = 0.25*n2+2.25*q+0.75*sqrt(n4+2.0*n2*q+9.0*q*q);
		}
	}
	
	return epsilon;
}

//Setup function for extended means function. LEJ 05/31/2009
//Creates a file to store extended means info in.
void setup_extended_means()
{
	Cnsts *cnsts = GetCnsts();
	string xmf = cnsts->fileout + int2str(cnsts->simnum) + "xdata.dat";
	FILE* f = fopen(xmf.c_str(), "wa");
	fprintf(f, "Sampling Frequency:\n%i\n", cnsts->xmfreq);
	fclose(f);
}

/*Extended means function, intended for finding a director axis in a simulation
 without an external E, e.g. SJB's liquid crystal simulations. Note that this
 function requires much more computational effort and produces significantly
 larger output files than the regular version. LEJ 05/31/2009*/
void get_extended_means(Molecule** BeOids)
{
	Cnsts *cnsts = GetCnsts();
	
	//Allocate memory and assume that the dipoles are aligned along the z asis of the molecules
	//This assumption may need to be changed later, but is currently used to make this algorithm
	//independant of dipole magnitude, allowing use with non-dipolar LC systems
	
	Vec3 tempcos, dipoleaxis(0.0,0.0,1.0), totalcos;
	Mat33 tenstemp, totaltens;
	totaltens.M3Zeros();
	
	//Loop over all dipoles in simulation
	for(int i = 0; i < cnsts->n_oids; i++)
	{
		//Once again, solvent molecules can be excluded from the average
		if (BeOids[i]->averageme == 1)
		{
			//Calculate cx, cy, cz
			tempcos = BeOids[i]->rot*dipoleaxis;
			totalcos += tempcos;
		
			//calculate c2 tensor
			TensorProduct3(tempcos, tempcos, tenstemp);
			totaltens += tenstemp;
		}
	}
	
	Vec3 meancos = totalcos/cnsts->n_avg;
	Mat33 cos2tens = totaltens/cnsts->n_avg;
	
	//Print cos vector and cos2 tensor to file
	string xmf = cnsts->fileout + int2str(cnsts->simnum) + "xdata.dat";
	FILE* f = fopen(xmf.c_str(), "a");	
	fprintf(f,"\n\n%s\n\n",meancos.V3Str().c_str());
	fprintf(f,"%s\n",cos2tens.M3Str().c_str());
			
	//clean up
	fclose(f);
}

//Setup function for extended means function. LEJ 05/31/2009
//Creates a file to store extended means info in.
void setup_binary_means()
{
	Cnsts *cnsts = GetCnsts();
	string xmfa = cnsts->fileout + int2str(cnsts->simnum) + "xdataa.dat";
	string xmfb = cnsts->fileout + int2str(cnsts->simnum) + "xdatab.dat";
	FILE* f = fopen(xmfa.c_str(), "wa");
	FILE* g = fopen(xmfb.c_str(), "wa");
	fprintf(f, "Type A, Sampling Frequency:\n%i\n", cnsts->xmfreq);
	fprintf(g, "Type B, Sampling Frequency:\n%i\n", cnsts->xmfreq);
	fclose(f);
	fclose(g);
}


/*Extended means function for binary systems, outputs separate A and B xdata files.
 Note that this function requires much more computational effort and produces significantly
 larger output files than the regular version. LEJ 04/01/2011*/
void get_binary_means(Molecule** BeOids)
{
	Cnsts *cnsts = GetCnsts();
	
		//Allocate memory and assume that the dipoles are aligned along the z asis of the molecules
		//This assumption may need to be changed later, but is currently used to make this algorithm
		//independant of dipole magnitude, allowing use with non-dipolar LC systems
	
	Vec3 tempcos, dipoleaxis(0.0,0.0,1.0);
	Vec3* totalcos = new Vec3[2];
	Mat33 tenstemp;
	Mat33* totaltens = new Mat33[2];
	totaltens[0].M3Zeros();
	totaltens[1].M3Zeros();
	
		//Loop over all dipoles in simulation
	for(int i = 0; i < cnsts->n_oids; i++)
	{
		//Calculate cx, cy, cz
		tempcos = BeOids[i]->rot*dipoleaxis;
			
		//calculate c2 tensor
		TensorProduct3(tempcos, tempcos, tenstemp);

		totalcos[BeOids[i]->mytype] += tempcos;
		totaltens[BeOids[i]->mytype] += tenstemp;
	}
	
	Vec3 meancosA = totalcos[0]/cnsts->n_avg;
	Mat33 cos2tensA = totaltens[0]/cnsts->n_avg;
	Vec3 meancosB = totalcos[1]/cnsts->n_avg;
	Mat33 cos2tensB = totaltens[1]/cnsts->n_avg;
	
	//Print cos vector and cos2 tensor to file
	string xmfa = cnsts->fileout + int2str(cnsts->simnum) + "xdataa.dat";
	string xmfb = cnsts->fileout + int2str(cnsts->simnum) + "xdatab.dat";
	FILE* f = fopen(xmfa.c_str(), "a");
	FILE* g = fopen(xmfb.c_str(), "a");
	fprintf(f,"\n\n%s\n\n",meancosA.V3Str().c_str());
	fprintf(f,"%s\n",cos2tensA.M3Str().c_str());
	fprintf(g,"\n\n%s\n\n",meancosB.V3Str().c_str());
	fprintf(g,"%s\n",cos2tensB.M3Str().c_str());
	
	//clean up
	fclose(f);
	fclose(g);
	delete[] totalcos;
	delete[] totaltens;
}


//Final dielectric constant calculation using mean values LEJ 07/13/2009
//Declared as void in order to easily pass both eps and eps_s back w/o making a new class.
//This should probably be changed to use a typedef or something - LEJ 02/17/2010
//Inputs:
//	M - average total dipole moment vector
//	M2 - mean squared total dipole moment vector
//	eps_array - passthrough for returning average epsilon and epsilon_s (reaction field dielectric constant)
void final_eps(Vec3 &M, double M2, double* eps_array)
{
	Cnsts *cnsts = GetCnsts();
	
	//Allocate memory
	double epsilon;
	double eps_s = cnsts->epsilon;
	double q, qs;
	double n2 = cnsts->n2;
	double n4 = n2*n2;
	
	//Dipole moment scaling factor
	double y_scale = 4*pi/(9*cnsts->V*cnsts->kT);
		
	//Calculate scalar dielectric constant based on external field
	if (cnsts->Efield != 0)
	{
		//Constant external potential (E)
		if (cnsts->realfield == 1)
		{
			epsilon = 4.0*pi*M.vec[2]/(cnsts->Efield*cnsts->Escale*cnsts->V) + cnsts->n2;
			if (cnsts->scrf == 1) eps_s = epsilon;
		}
		//Constant effective/internal field (E0)
		else 
		{
			q = (4.0*pi*M.vec[2])/(3.0*cnsts->Efield*cnsts->Escale*cnsts->V);
			if (cnsts->scrf == 0) epsilon = (6.0*q*eps_s+2.0*n2*eps_s+n4)/(n2+2.0*eps_s-3.0*q);
			else
			{
				epsilon = 0.25*n2+2.25*q+0.75*sqrt(n4+2.0*n2*q+9.0*q*q);
				eps_s = epsilon;
			}
		}

	}
	
	//Calculate dielectric constant from variance of total dipole moment (Kirkwood method)
	else
	{
		q = y_scale*(M2-M*M); //calculate product of dipole density and Var(M)
		if (cnsts->scrf == 1) //Calculate average epsilon_s from mean square M
		{
			qs = y_scale*M2;
			eps_s = 0.25*n2+2.25*qs+0.75*sqrt(n4+2.0*n2*qs+9.0*qs*qs);
		}
		epsilon = (6.0*q*eps_s+2.0*n2*eps_s+n4)/(n2+2.0*eps_s-3.0*q);
	}
	
	//Concantenate into an array. For an isotropic phase, these numbers should be nearly equal if converged
	eps_array[0] = epsilon; eps_array[1] = eps_s;
}
