/*
 *
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *
 *  get_rs_mus.cpp
 *  Created by ljohnson on 12/30/08.
 *	Distance calculation, dipole rotation, and helper functions
 */

#include "get_rs_mus.h"

//Distances between particles in different geometries
//Inputs:
//BeOids - the array of all of the molecules in the system
//dr - position of the particle the distances are being calculated relative to
//r_mus - the vector of current positions of molecules
void get_distances(Molecule** BeOids, Vec3 &dr, Vec3* r_mus)
{
	Cnsts *cnsts = GetCnsts();
	for(int i = 0; i < cnsts->n_oids; i++)
	{
		r_mus[i] = BeOids[i]->center - dr;
		apply_pbcs(r_mus[i]);
	}
}

//Periodic boundary condition helper function
//rvec is the distance vector having the PBC correction applied. See Allen and Tildesley for details.
void apply_pbcs(Vec3& rvec)
{
	Cnsts *cnsts = GetCnsts();
	
	//Rectangular box (PBCs on arbitrary axes)
	if (cnsts->latticetype < 2)
	{
		for(int j = 0; j < 3; j++)
		{
			//Apply PBC correction if PBCs are used along that axis
			if (cnsts->PBCs[j] == 1)
			{
				double pbsadjust = rvec.vec[j]/cnsts->boxlength[j];
				rvec.vec[j] -= cnsts->boxlength[j]*qround(pbsadjust);
			}
		}
	}
	
	//Spherical box (PBCs not yet supported)
	else if (cnsts->latticetype == 3) return;
	
	//Cylindrical box (optional PBCs on z-axis only)
	else if (cnsts->latticetype == 4)
	{
		if (cnsts->PBCs[2] == 1)
		{
			double pbsadjust = rvec.vec[2]/cnsts->boxlength[2];
			rvec.vec[2] -= cnsts->boxlength[2]*qround(pbsadjust);
		}
	}
	
	//Default condition
	else
	{
		printf("Lattice type not recognized. Exiting.");
		exit(1);
	}
}


//Rotate dipoles from particle frame into lab frame at start of simulation
//Inputs:
//	BeOids - the array of all molecules in the system
//	mus - array of lab-frame dipole moments (passthrough - actually an output)
void get_dipoles(Molecule** BeOids, Vec3* mus)
{
	Cnsts *cnsts = GetCnsts();
	
	//Loop over all dipoles in simulation
	for(int i = 0; i < cnsts->n_oids; i++)
	{
		mus[i] = BeOids[i]->rot*BeOids[i]->ellipse->dipole;
	}		
}


//Update only the dipole that was moved
//Inputs:
//	TheOid - The active molecule
//	themu - The dipole moment (in mus) being updated (passthrough)
void update_dipoles(Molecule* TheOid, Vec3 &themu)
{
	//This is just the above function without the loop.
	themu = TheOid->rot*TheOid->ellipse->dipole;
}


//Off-center dipole correction
//This is a slow, temporary algorithm intended as a stopgap until rotated dipole vectors are directly stored
void offcenter_mu(Molecule** BeOids, Vec3 &tmprmu, int i, int k)
{
	if (i != k)
	{
		//Correct 'it' dipole
		Vec3 mycenter = BeOids[k]->center + (BeOids[k]->rot*BeOids[k]->ellipse->mu_pos); //fixed by AT
	
		//Correct other dipole
		Vec3 thecenter = BeOids[i]->center + (BeOids[i]->rot*BeOids[i]->ellipse->mu_pos); //fixed by AT
	
		//Calculate difference between vectors
		tmprmu = thecenter-mycenter;
	}
	else tmprmu.V3Zeros();
}

