/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *
 *  touch.cpp
 *  Created by robin on 2/3/09.
 */

#include "newtouch.h"

/*Touch, an algorithm designed for determining the closest contact distance between two ellipsoids at arbitrary angles,
reducing the three-dimensional problem to a one-dimensional problem in terms of a scalar varaible lambda, then iteratively
solving a system of linear equations to optimize lambda. For more info see Perram & Wertheim, J. Comp. Phys. 58 409-416 (1985)
Initially written in Matlab by BHR, ported to C++ by RB in 02/09, then harmonized with current codebase by LEJ in 04/09,
and optimized by LEJ in 01/10. 
Inputs:
 BeOids - the entire ensemble of molecules
 kk - the index of the currently active molecule
 i - the index of the molecule it is interacting with
 Rmu - the distance vector between the centers of the molecules
 distance - the norm of the distance between molecules
Return value is the effective LJ sigma for the molecules at their current positions*/
double touch(Molecule** BeOids,int &kk,int &i, Vec3 &Rmu, double &distance)
{
	//Normalize distance vector
	//t is Rmu, nt is distance
	double invdist = 1.0/distance;
	double invdist2 = invdist*invdist;
	Vec3 z =Rmu*invdist;

	//Minimization vectors
	Vec3 V, VtempX;
	
	//Create A and B matrices, 
	Mat33 A;//3x3 with saxes for the kkth oid on the diagonal.
	Mat33 B;//3x3 with saxes for the ith oid on the diagonal.

	for (int j=0;j<3;j++)
	{
		A.mat[j][j]=(BeOids[kk]->ellipse->saxes.vec[j]*BeOids[kk]->ellipse->saxes.vec[j]*invdist2);
		B.mat[j][j]=(BeOids[i]->ellipse->saxes.vec[j]*BeOids[i]->ellipse->saxes.vec[j]*invdist2);
	}
	
	//Create lab frame version of A and B matrices.
	Mat33 A_LF = BeOids[kk]->rot*A*BeOids[kk]->rot.M3Transpose();
	Mat33 B_LF = BeOids[i]->rot*B*BeOids[i]->rot.M3Transpose();
	
	//Create CI matrix and its inverse.
	Mat33 CI, CIinv;
	
	//Flag for singular matrices
	bool sing = 0;

	double lamx = 0.5;
	double Var = 1.0; // trying different forms found 6 loops gives 7 figs for Fx
	while (Var > 0.000001) //loop until variance in distance is sufficiently small
	{
		//Populate CI matrix
		double lamoid = lamx;
		for (int j=0;j<3;j++)
		{
			CI.mat[j][0] = lamx*B_LF.mat[j][0]-lamx*A_LF.mat[j][0]+A_LF.mat[j][0];
			CI.mat[j][1] = lamx*B_LF.mat[j][1]-lamx*A_LF.mat[j][1]+A_LF.mat[j][1];
			CI.mat[j][2] = lamx*B_LF.mat[j][2]-lamx*A_LF.mat[j][2]+A_LF.mat[j][2];
		}
		
		//Invert CI matrix
		CIinv = CI.M3Inv(sing);
		
		if (sing == 1)
		{
			printf("Distance calculation failed. i = %i, kk = %i, lamx = %f\n", i, kk, lamx);
			printf("CI matrix is:\n%s\n",CI.M3Str().c_str());
			double dettemp = CI.M3Det();
			printf("Det(CI) is:%f\n", dettemp);
			printf("CI' is:\n%s\n",CIinv.M3Str().c_str());
			exit(2);
		}
		
		//Solve z = CI*V for V using inverse
		V = CIinv*z;
		VtempX = V*A_LF;
		double numerator = VtempX*V;
		VtempX = V*B_LF;
		double denominator = VtempX*V;
		
		if (denominator == 0)
		{
			printf("Denominator = 0 in touch between oids %i and %i",kk,i);
			exit(3);
		}
		
		//Calculate minimization parameter lambda
		double xlx = sqrt(numerator/denominator);
		lamx = xlx/(1.0+xlx);
		double Var_temp = lamx-lamoid;
		Var = Var_temp*Var_temp;
	}
	
	//Reconstruct CI and run a final iteration once converged
	for (int j=0;j<3;j++)
	{
		CI.mat[j][0] = lamx*B_LF.mat[j][0]-lamx*A_LF.mat[j][0]+A_LF.mat[j][0];
		CI.mat[j][1] = lamx*B_LF.mat[j][1]-lamx*A_LF.mat[j][1]+A_LF.mat[j][1];
		CI.mat[j][2] = lamx*B_LF.mat[j][2]-lamx*A_LF.mat[j][2]+A_LF.mat[j][2];
	}
	
	CIinv = CI.M3Inv(sing);
	if (sing == 1)
	{
		printf("Distance calculation failed. i = %i, kk = %i, lamx = %f\n", i, kk, lamx);
		printf("CI matrix is:\n%s\n",CI.M3Str().c_str());
		double dettemp = CI.M3Det();
		printf("Det(CI) is:%f\n", dettemp);
		printf("CI' is:\n%s\n",CIinv.M3Str().c_str());
		exit(2);
	}
	
	V = CIinv*z;
	
	//Calculate inverse of contact distance. Crash if it is zero.
	double Fx = lamx*(1.0-lamx)*(z*V);
	if (Fx==0)
	{
		printf("Fx is zero in touch between oids %i and %i.\n",kk,i);
		exit(3);
	}
	
	//Bomb out if molecules are overlapped. This shouldn't happen, except very minorly
	//and can generally be kept commented out.
	/*if ((Fx<1.0)&&(distance>16.0)){
		printf("distance is%f\n",distance);
		printf("Fx is less than 1.0 in touch%f\n",Fx);
		exit(1);
	}*/
	
	double Rx = 1.0/Fx;
	return Rx;
}
