/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *
 *  output_data.cpp
 *	Outputs data from MC simulations to a file, with a summary sent to standard out. Also creates a terse, machine-readable
 *	summary of output from jobs in a batch for batch size > 1. Requires the arrays of cos^n moments and total dipole for
 *	every cycle of the simulation, as well as the simulation start time and the integer array containing the acceptance rates.
 *  Created by ljohnson on 10/9/08, Updated 01/16/2011.
 */

#include "output_data.h"

//Output header. This prints the constants in advance to allow you to check if you are running the right sim.
void output_header()
{
	Cnsts *cnsts = GetCnsts();
	
	//Allocate strings
	string ptypes, rtypes, ttypes, ltype, mmstatus, scstatus, simtype,PBCtype, vdwtype, thefilename, Summfn, fieldstr, cfn;
	
	//generate the filename
	thefilename = cnsts->fileout + int2str(cnsts->simnum) + ".dat";
	
	//Calculate reduced dipole strength y
	double mzA = cnsts->muzA*cnsts->mucorr;
	double mzB = cnsts->muzB*cnsts->mucorr;
	double y = 4*pi*((double)(cnsts->num_A)*mzA*mzA + (double)(cnsts->num_B)*mzB*mzB)/(9*cnsts->V*cnsts->kT);
	
	//Record whether particle types are averaged, and whether they can rotate and translate
	if (cnsts->averageA == 1) ptypes += "A ";
	if (cnsts->averageB == 1) ptypes += "B";
	if (cnsts->stillmembersA == 0)
	{
		rtypes += "A ";
		if (cnsts->rot_notransA == 0) ttypes += "A ";
	}
	if (cnsts->stillmembersB == 0)
	{
		rtypes += "B";
		if (cnsts->rot_notransB == 0) ttypes += "B";
	}
	
	//Check which loader was used
	if (cnsts->loader == 1) ltype += "SEQUENTIAL";
	else if (cnsts->loader == 2) ltype += "RANDOM";
	else if (cnsts->loader == 3)
	{
		ltype += "FILE";
		cfn = cnsts->filein + int2str(cnsts->simnum) + "chk" + cnsts->chkpt + "coords.dat";
	}
	
	
	//Obtain info about geometry, cutoffs, PBCs, and vdW model
	if(cnsts->cut == 1)
	{
		scstatus = "SPHERICAL";
	}
	else scstatus = "NONE";
	
	PBCtype = "axes: ";
	if (cnsts->latticetype == (1 || 2))
	{
		simtype = "CUBIC";
		if (cnsts->PBCs[0] == 1) PBCtype += "X ";
		if (cnsts->PBCs[1] == 1) PBCtype += "Y ";
		if (cnsts->PBCs[2] == 1) PBCtype += "Z";
	}
	else if (cnsts->latticetype == 3)
	{
		simtype = "ISOLATED_SPHERE";
		if (cnsts->PBCs[0] == 1) PBCtype = "R";
	}
	else if (cnsts->latticetype == 4)
	{
		simtype = "CYLINDER";
		if (cnsts->PBCs[0] == 1) PBCtype += "R ";
		if (cnsts->PBCs[2] == 1) PBCtype += "Z";
	}
	
	if (cnsts->vdwtype == 1) vdwtype = "SIMPLE";
	else if (cnsts->vdwtype == 2) vdwtype = "EXTENDED";
	else if (cnsts->vdwtype == 3) vdwtype = "SOFT SPHERE";
	else if (cnsts->vdwtype == 4) vdwtype = "SIMPLE TOUCH";
	else if (cnsts->vdwtype == 5) vdwtype = "SOFT SPHERE TOUCH";
	else vdwtype = "EXTENDED";
	
	//Obtain info about external field model
	if (cnsts->Efield != 0)
	{
		if (cnsts->realfield == 1) fieldstr = "defined by the POLING VOLTAGE";
		else fieldstr = "defined as the CAVITY FIELD";
	}
	else fieldstr = "DISABLED";
	
	//get timestamp for file
	time_t rawtime;
	struct tm * timeinfo;
	time ( &rawtime );
	timeinfo = localtime ( &rawtime );
	
	//start output block
	FILE* f = fopen(thefilename.c_str(),"wa");
	fprintf(f, "Robinson Group Classical Monte Carlo Molecular Simulator.\nDevelopment build %s.\n", BUILD);
	if (cnsts->test == 1) fprintf(f,"DIAGNOSTIC MODE ON -- RNG LOCKED\n");
	else fprintf(f,"\n");
	if (cnsts->vdwtype == 4 || cnsts->vdwtype == 5) fprintf(f, "TOUCH ENABLED\n");
	else fprintf(f, "\n");
	fprintf(f, "Output file stamped on %s \n\nSimulation parameters:\n\n", asctime(timeinfo));
	fprintf(f, "Simulation length was: %li\n", cnsts->steps);
	fprintf(f, "Simulation type was: %s\n", simtype.c_str());
	fprintf(f, "RNG seed was: %li\n",cnsts->rngseed);
	fprintf(f, "Steps with only vdw interactions: %li\n",cnsts->randsteps);
	fprintf(f, "Steps to be used for averaging: %li\n",cnsts->avg_steps);
	fprintf(f, "Particle types averaged over for <cos^n(theta)> and correlation functions: %s\n", ptypes.c_str());
	fprintf(f, "Particle types allowed to translate: %s\n", ttypes.c_str());
	fprintf(f, "Particle types allowed to rotate: %s\n", rtypes.c_str());
	fprintf(f, "Maximum move sizes were: %f A (trans)  %f rad (rot)\n", cnsts->maxtrans, cnsts->maxrot);
	fprintf(f, "Coordinates were loaded using %s loader.",ltype.c_str());
	if (cnsts->loader == 3) fprintf(f, "Source file was %s\n", cfn.c_str());
	else fprintf(f,"\n");
	fprintf(f, "PBCs were used on the following %s\n", PBCtype.c_str());
	fprintf(f, "Cutoff type was	%s. ", scstatus.c_str());
	if (scstatus != "NONE") fprintf(f, "Cutoff distances were: %f A (LJ)  %f A (electrostatic)\n", cnsts->ljcut, cnsts->escut);
	else fprintf(f, "\n");
	fprintf(f, "Van Der Waals type: %s\n", vdwtype.c_str());
	if (cnsts->vdwtype == 4 || cnsts->vdwtype == 5) fprintf(f, "Touch cutoff was: %f (A)\n",cnsts->tcut);
	else fprintf(f, "Touch was not used.\n");
	fprintf(f, "External field was %s\n", fieldstr.c_str());
	if (cnsts->anneal == 1) fprintf(f, "Simulated annealing was used during first half of randomization. Tmax was %f\n", cnsts->T + (double)(cnsts->randsteps/10));
	else fprintf(f, "Simulated annealing was not used.\n");
	fprintf(f, "\nT (K):\n%f\n",cnsts->T);
	fprintf(f, "\nNumber of particles:\n%i\n",cnsts->n_oids);
	fprintf(f, "\nInitial lattice dimensions were:\n%i	%i	%i\n", cnsts->lin_dim[0], cnsts->lin_dim[1], cnsts->lin_dim[2]);
	fprintf(f, "\nStarting nearest-neighbor distances (A) were:\n%f\n", cnsts->nndist);
	if (simtype == "ISOLATED_SPHERE") fprintf(f, "\nSimulation cell radius (A):\n%f\n", cnsts->spherecylr);
	else if (simtype == "CYLINDER") fprintf(f, "\nSimulation cell r and z (A):\n%f	%f\n", cnsts->spherecylr, cnsts->boxlength[2]);
	else fprintf(f, "\nLength of each side of box (A):\n%f	%f	%f\n", cnsts->boxlength[0], cnsts->boxlength[1], cnsts->boxlength[2]);
	fprintf(f, "\nSimulation cell volume (A^3):\n%f\n", cnsts->V);
	fprintf(f, "\nNumber of A and B oids:\n%i	%i\n", cnsts->num_A, cnsts->num_B);
	fprintf(f, "\nDimensions of A oids (A):\n%f	%f	%f\n", cnsts->r0A[0], cnsts->r0A[1], cnsts->r0A[2]);
	fprintf(f, "\nDimensions of B oids (A):\n%f	%f	%f\n", cnsts->r0B[0], cnsts->r0B[1], cnsts->r0B[2]);
	fprintf(f, "\nGas-phase dipole moments of A and B oids (D):\n%f	%f\n", cnsts->muzA, cnsts->muzB);
	fprintf(f, "\nDipole offsets of A oids (A):\n%f	%f	%f\n", cnsts->m0A[0], cnsts->m0A[1], cnsts->m0A[2]);
	fprintf(f, "\nDipole offsets of B oids (A):\n%f	%f	%f\n", cnsts->m0B[0], cnsts->m0B[1], cnsts->m0B[2]);
	fprintf(f, "\nVdW strength of A and B oids (pico-ergs):\n%f	%f\n", cnsts->vdwA, cnsts->vdwB);
	fprintf(f, "\nExternal field strength (V/um):\n%f\n", cnsts->Efield);
	fprintf(f, "\nReduced dipole strength y (dimensionless):\n%f\n",y);
	if (cnsts->scrf == 0) fprintf(f, "\nReaction field dielectric constant (dimensionless):\n%f\n", cnsts->epsilon);
	else fprintf(f, "\nDynamic reaction field used. Initial dielectric constant was:\n%f\n", cnsts->epsilon);
	fprintf(f, "\nContribution to epsilon from polarizable medium (n^2, dimensionless):\n%f\n", cnsts->n2);
	if (cnsts->polarize ==1 ) fprintf(f, "\nOnsager polarizability active. Corrected dipole moments (D) are:\n%f	%f\n", cnsts->muzA*cnsts->mucorr, cnsts->muzB*cnsts->mucorr);
	fprintf(f, "\nStarting simulation on %s\n\n\n", asctime(timeinfo));
	fclose(f);
}

//Write saved data for a run to a log file
//Inputs:
//	cosMeans - array of vectors containing order parameters from all cycles
//	bigMs - array of vectors containing total dipole moments from each cycle
//	totalV - array of vectors containing total energies from each cycle
//	tstart - start time for run, used for calculating run time
//	accepted - array of counts of accepted moves
//	moved - array of total squared displacements
//	Step - the current step, counting from 1 (not 0!)
void output_data(Vec3* cosMeans, Vec3* bigMs, Vec3* totalV, double tstart, long* accepted, double* moved, long Step)
{
	Cnsts *cnsts = GetCnsts();
	
	//Allocate memory. Not done elegantly, but everything here is local.
	long i;
	double temptotU = 0.0, totalU2 = 0.0;
	string thefilename, Summfn;
	Vec3 tmpcos, totalM, tempUU, expecU, expecU2, expecUU, acceptance, Apercent;
	Mat33 M2tenstemp(0.0), totalM2tens(0.0), MMtenstemp(0.0), eps_tens(0.0);
	
	//Generate the filename
	thefilename = cnsts->fileout + int2str(cnsts->simnum) + ".dat";
	
	//Calculate acceptance rate and determine if mu-mus were on. Note that the argument "Step" is actually
	//splus1 for consistancy with avgsteps and startavg.
	acceptance.vec[0] = (double)accepted[0]/(double)(cnsts->randsteps*cnsts->n_oids);
	acceptance.vec[1] = (double)accepted[1]/(double)((cnsts->steps-cnsts->randsteps-cnsts->avg_steps)*(cnsts->n_oids));
	acceptance.vec[2] = (double)accepted[2]/(double)(cnsts->avg_steps*cnsts->n_oids);
	Apercent = acceptance*100.0;
	
	//Precalculate or type-convert some useful values
	long startavg = cnsts->steps - cnsts->avg_steps;	//Step at which averaging starts
	long avgsteps = Step - startavg;
	double avgd = (double)avgsteps;
	double mzA = cnsts->muzA*cnsts->mucorr; double nA = cnsts->num_A;
	double mzB = cnsts->muzB*cnsts->mucorr; double nB = cnsts->num_B;
	double nTot = nA + nB;
	double NkT = (double)(cnsts->n_oids)*cnsts->kT;
	double iNkT = 1.0/NkT;
	


	//Get timestamp for file
	time_t rawtime;
	struct tm * timeinfo;
	time ( &rawtime );
	timeinfo = localtime ( &rawtime );
	
	//Open file for appending data
	FILE* f = fopen(thefilename.c_str(),"a");
	fprintf(f, "Data processed at %s", asctime(timeinfo));

	//Abort if there is no data to append
	if ((Step - avgsteps) < 1)
	{
		printf("\nNo data available to average!\n");
		fprintf(f, "No data available to average!\n");
		fclose(f);
		return;
	}
	else fprintf(f, "Expectation values will be calculated from final %li steps.\n", avgsteps);
	
	//Average moments
	fprintf(f, "\nOutput data:\n\nMoments for all %li cycles:\n\ncos		cos^2		cos^3\n", Step);
	for (i = 0; i < Step; i++)
	{
		fprintf(f,"%s\n",cosMeans[i].V3Str().c_str());
		if (i >= startavg) tmpcos+=cosMeans[i];
	}
	
	Vec3 expeccos = tmpcos/avgd;

	fprintf(f, "\n<cos>		<cos^2>		<cos^3>\n%s\n", expeccos.V3Str().c_str());
	printf("\n<cos^n> = %s\n",expeccos.V3Str().c_str());
	
	//Variance and skew of <cos>
	double var_cos = expeccos.vec[1]-expeccos.vec[0]*expeccos.vec[0];
	double skew_cos = expeccos.vec[2]-3*expeccos.vec[0]*expeccos.vec[1]+2*expeccos.vec[0]*expeccos.vec[0]*expeccos.vec[0];
	fprintf(f,"\nCentral moments of <cos>\n");
	fprintf(f,"\nVariance:\n%f\n", var_cos);
	fprintf(f,"\nSkewness:\n%f\n", skew_cos);
	
	//Legendre polynomials of <cos>
	double P2 = (3*expeccos.vec[1]-1)/2;
	double P3 = (5*expeccos.vec[2]-3*expeccos.vec[0])/2;
	fprintf(f,"\n<P2>:\n%f\n", P2);
	fprintf(f,"\n<P3>:\n%f\n", P3);
	
	//Calculate <M> and the <M2> tensor
	fprintf(f, "\n\n\nM values for all %li cycles (Debye):\n\nMx		My		Mz\n", Step);
	for (i = 0; i < Step; i++)
	{
		fprintf(f, "%s\n", bigMs[i].V3Str().c_str());
		
		if(i >= startavg) //calculate tensor only if in equilibrum
		{
			totalM+=bigMs[i];
			TensorProduct3(bigMs[i], bigMs[i], M2tenstemp);
			totalM2tens += M2tenstemp;
		}
	}
	
	Vec3 expecM = totalM/avgd;
	Vec3 MoN = expecM/nTot;
	printf("<M> = %s\n", expecM.V3Str().c_str());
	printf("<M>/N = %s\n", MoN.V3Str().c_str());
	fprintf(f, "\n<Mx>		<My>		<Mz>\n%s\n",expecM.V3Str().c_str());
	
	//Calculate <M^2> and the <M2> tensor
	Mat33 expecM2tens = totalM2tens/avgd;
	double expecM2 = expecM2tens.M3Trace();
	printf("<M2> = %f\n", expecM2);
	fprintf(f, "\n<M2> tensor:\n");
	fprintf(f,"%s\n",expecM2tens.M3Str().c_str());
	
	fprintf(f, "\n<M2>\n%f\n", expecM2);
	
	//Calculate <M>^2
	double expecMM = expecM*expecM;
	printf("<M>2 = %f\n", expecMM);
	fprintf(f, "\n<M>2\n%f\n", expecMM);
	
	//Calculate covariance matrix
	TensorProduct3(expecM, expecM, MMtenstemp);
	Mat33 covar = expecM2tens - MMtenstemp;
	
	fprintf(f, "\nCovariance matrix:\n");
	fprintf(f,"%s\n",covar.M3Str().c_str());
	
	//Calculate variance of M
	double var_M = expecM2-expecMM;
	fprintf(f, "\nVar(M):\n%f\n", var_M);
																					
	//Calculate scalar Kirkwood g-factors
	double gdenom = (double)cnsts->averageA*nA*qpwr(mzA,2) + (double)cnsts->averageB*nB*qpwr(mzB,2);
	double g = expecM2/gdenom;
	double gfull = (expecM2 - expecMM)/gdenom;
	printf("g_M2 = %f\n",g);
	fprintf(f, "\nKirkwood g-factors:\n\ng_M2:\n%f\n", g);
	printf("g_Var = %f\n",gfull);
	fprintf(f, "\ng_Var:\n%f\n", gfull);
	
	//Calculate epsilon (and <epsilon_s> (RF dielectric constant) if SCRF was used)
	double epsilon[2];
	final_eps(expecM, expecM2, epsilon);
	printf("epsilon = %f\n",epsilon[0]);
	fprintf(f, "\nScalar dielectric constant:\n%f\n", epsilon[0]);
		//fprintf(f, "\nScalar dielectric constant:\n%f\n\nDielectric tensor (exp. feature--not yet trustworthy):\n", epsilon);
		//fprintf(f, "%s", eps_tens.M3Str().c_str());
	
	//Calculate electric field strength in medium using Lorentz-Onsager cavity approximation if vacuum field is used.
	//If field in medium is used, calculate vacuum field strength.
	//See Onsager (JACS, 1936), March & Tosi, Atomic Dynamics in Liquids, and Jackson, Classical Electrodynamics for details. - LEJ 09/02/2009
	double Ecorr,Eint,Eext;
	if (cnsts->scrf == 1) Ecorr = (2.0*epsilon[0] + cnsts->n2)/(3.0*epsilon[0]);
	else Ecorr = (2.0*cnsts->epsilon + cnsts->n2)/(2.0*cnsts->epsilon + epsilon[0]);
	if (cnsts->realfield == 1)
	{
		Eext = cnsts->Efield;
		Eint = Eext/Ecorr;
		fprintf(f, "\nZ-component of electric field in cavity (V/um):\n%f\n", Eint);
	}
	else
	{
		Eint = cnsts->Efield;
		Eext =Eint*Ecorr;
		fprintf(f, "\nZ-component of electric field in bulk media (V/um):\n%f\n", Eext);
	}
	
	//Display total energy
	fprintf(f,"\n\n\nPotential Energy values for all %li cycles divided by kT:\n",Step);
	fprintf(f,"\nmuE		mumu		VdW\n");
	for (i = 0; i < Step; i++)
	{
		fprintf(f,"%s\n",(totalV[i]*iNkT).V3Str().c_str());
		if (i >= startavg)
		{
			expecU += totalV[i];
			temptotU = totalV[i].V3Sum();
			Vec3Amult(totalV[i], totalV[i], tempUU);
			expecU2 += tempUU;
			totalU2 += temptotU*temptotU;
		}
	}
	
	//Calculate variance of energy
	expecU /= avgd;
	expecU2 /= avgd;
	totalU2 /= avgd;
	Vec3Amult(expecU, expecU, expecUU);
	Vec3 VarU = expecU2 - expecUU;
	double totalU = expecU.V3Sum();
	double VarTotalU = totalU2-totalU*totalU;
	
	//Heat capacity. Note that this ONLY includes intermolecular potential (kB), kinetic (3/2 kB), and rotational contributions.
	double Cv = (VarTotalU/(nTot*cnsts->kB*cnsts->T*cnsts->T) + 2.5*cnsts->kB)*cnsts->perg2Jmol;	//Output in J/mol-K
	
	//Display thermodynamic parameters in dimensionless (NkT) units unless otherwise noted
	fprintf(f,"\n<UmuE/NkT>	<Umumu/NkT>	<ULJ/NkT>\n%s\n", (expecU/NkT).V3Str().c_str());
	fprintf(f, "\nVar(UmuE)/NkT	Var(Umumu)/NkT	Var(ULJ)/NkT\n%s\n", (VarU/(NkT*NkT)).V3Str().c_str());
	printf("<U/NkT> = %s\n", (expecU/NkT).V3Str().c_str());
	printf("Var(U/NkT) = %s\n", (VarU/(NkT*NkT)).V3Str().c_str());
	printf("<Utotal/NkT> = %f\n", totalU/NkT);
	fprintf(f,"\n<Utotal/NkT>\n%f\n", totalU/NkT);
	fprintf(f,"\nVar(Utotal/NkT)\n%f\n", VarTotalU/(NkT*NkT));
	fprintf(f,"\n<Utotal> (pErg)\n%f\n", totalU);
	printf("dEvap (kJ/mol) = %f\n", -0.001*cnsts->perg2Jmol*totalU/nTot);
	printf("Cv (J/mol K) = %f\n",Cv);
	fprintf(f,"\ndEvap (kJ/mol)\n%f\n", -0.001*cnsts->perg2Jmol*totalU/nTot);
	fprintf(f,"\nCv (J/mol K)\n%f\n",Cv);
	
	//Calculate and display RMS displacement
	moved[0] = sqrt(moved[0]/(double)Step);
	moved[1] = sqrt(moved[1]/(double)Step);
	printf("RMS move sizes = %f	%f\n", moved[0], moved[1]);
	fprintf(f, "\n\n\nRMS move sizes:\n\ntrans (A)	rot (rad)\n%f	%f\n", moved[0], moved[1]);
	
	//Display acceptance rate
	printf("Acceptance rates are (%%): %s\n",Apercent.V3Str().c_str());
	fprintf(f, "\nAcceptance rates (%%):\n\nRandomize	PreEq		Equilibrium\n%s\n",Apercent.V3Str().c_str());
	
	//Display execution time - LEJ
	double endtime = clock();
	double tfp = (endtime-tstart)/((double)CLOCKS_PER_SEC);
	printf("Execution time: %f seconds.\n", tfp);
	fprintf(f, "\nExecution time: %f seconds.\n\nRun completed successfully.\n", tfp);
	fclose(f);
	
	//Generate summary file
	//Uses new (02/17/10) summary file format which properly handles multi-component systems
	Summfn = cnsts->fileout + "summary.dat";
	FILE * summ = fopen(Summfn.c_str(), "a");
	fprintf(summ, "\n%i\n", cnsts->simnum);
	fprintf(summ, "COS^N	%s\n", expeccos.V3Str().c_str());
	fprintf(summ, "M	%s\n",expecM.V3Str().c_str());
	fprintf(summ, "M2	%f\n",expecM2);
	fprintf(summ, "GK	%f	%f\n", g, gfull);
	fprintf(summ, "E	%f	%f\n", Eint, Eext);
	fprintf(summ, "EPSILON	%f	%f\n", epsilon[0], epsilon[1]);
	fprintf(summ, "U/NKT	%f	%f\n", totalU/NkT, VarTotalU/(NkT*NkT));
	fprintf(summ, "ACCEPT	%s\n",Apercent.V3Str().c_str());
	fprintf(summ, "RUNTIME\t%f\t%f\n",tfp,cnsts->Rand_Time);
	fclose(summ);
}



