/*
 * © 2012, Lewis E. Johnson
 * Developed by Lewis E. Johnson and Robin Barnes, Robinson Group, University of Washington.
 * Released under University of Illinois/NCSA Open Source License. See LICENSE.txt for details.
 *  setup.cpp
 *  Created by ljohnson on 5/31/09.
 *	Setup functions for Robinson MC code. These are simply intended to clean up the formerly-bloated MCmain.cpp
 *	Anything that the user doesn't edit should be here unless it would require passing a lot of variables around.
 *	Anything the user edits should remain in MCmain.cpp
 */

#include "setup.h"

//List of physical constants. Since kB isn't likely to change in the near future, the user doesn't need to edit these.
void phys_cnsts()
{
	Cnsts* cnsts = GetCnsts();
	
	//Load physical constants
	cnsts->ec = 4.8;
	cnsts->ergc = 1.0e-12;
	cnsts->perg2Jmol = 60200;	//Conversion from pico-ergs to J/mol
	cnsts->Escale = 3.33e-5; //conversion from MV/m into pico-ergs/Debye
	cnsts->kB = 1.38e-4; //Boltzmann's constant in pico-ergs/K
	cnsts->trackchange = 0; //counter for changing E and T. Should always be 0 to start.
	cnsts->changesteps = 0; //steps where E and T are changed. Must be same length as arrays below.
	cnsts->kT = cnsts->kB*cnsts->T;
	cnsts->beta = 1.0/cnsts->kT;
}

//Load vdW parameters into cnsts
void setup_vdw()
{
	Cnsts* cnsts = GetCnsts();
	
	//Set up vdw cutoff
	double bltemp;
	if (cnsts->latticetype < 3)
	{
		bltemp = cnsts->boxlength[0];
		if (cnsts->boxlength[1] < bltemp) bltemp = cnsts->boxlength[1];
		if (cnsts->boxlength[2] < bltemp) bltemp = cnsts->boxlength[2];
		bltemp /= 2.0;
	}
	else if (cnsts->latticetype > 2)
	{
		bltemp = cnsts->spherecylr;
		if (cnsts->latticetype == 4)
		{
			if (cnsts->boxlength[2]/2.0 < bltemp) bltemp = cnsts->boxlength[2]/2.0;
		}
	}
	
	if (cnsts->ljcut > bltemp)
	{
		cnsts->ljcut = bltemp;
		printf("Warning: Lennard-Jones cutoff is too large. Resizing to half of the box length (%f)\n", bltemp);
	}
	if (cnsts->ljcut > cnsts->escut)
	{
		cnsts->ljcut = cnsts->escut;
		printf("Warning: Lennard-Jones must be smaller than electrostatics cutoff (%f). Resizing...\n", cnsts->escut);
	}
	
	//Calculate parameters for LJ correction.
	double m = (double)(cnsts->LJexp[0])/2.0;
	double n = (double)(cnsts->LJexp[1])/2.0;
	double cfshift = pow(n/m,1/(n-m)); //shift used with solvent parameters
	double cr = n/m;
	double cr_scale = pow(cr,cr/(cr-1.0))/(cr-1.0); //scaling coefficient (4 for 12-6 potential)
	cnsts->fshift = cfshift;
	cnsts->r = cr_scale;
	
	//Precalculate vdw epsilon and ro values, as well as average distances for if Touch is not used
	//Will need to be replaced with an arbitrary matrix if more than two types of particles are allowed
	double travgA = average(cnsts->r0A, 3);
	double travgB = average(cnsts->r0B, 3);
	cnsts->pre_sigma[0][0] = 2.0*travgA;
	cnsts->pre_sigma[0][1] = travgA + travgB;
	cnsts->pre_sigma[1][0] = cnsts->pre_sigma[0][1];
	cnsts->pre_sigma[1][1] = 2.0*travgB;
	cnsts->pre_eps[0][0] = cnsts->vdwA;
	cnsts->pre_eps[0][1] = sqrt(cnsts->vdwA*cnsts->vdwB);
	cnsts->pre_eps[1][0] = cnsts->pre_eps[0][1];
	cnsts->pre_eps[1][1] = cnsts->vdwB;
	
	double nnscale = 1.25; //Scaling factor for outer edge of nearest-neighbor shell.
	
	
	//Set up Touch if it is used
	if (cnsts->vdwtype == 4 || cnsts->vdwtype == 5)
	{
		//Warn about performance
		printf("Warning: Touch is enabled—performance will be slower.\n");
		
		//Precalculate whether or not touch is used for a given pair of particle types
		bool tA = !((cnsts->r0A[0] == cnsts->r0A[1]) && (cnsts->r0A[1] == cnsts->r0A[2]));
		bool tB = !((cnsts->r0B[0] == cnsts->r0B[1]) && (cnsts->r0B[1] == cnsts->r0B[2]));
		cnsts->pre_touch[0][0] = tA;
		cnsts->pre_touch[0][1] = tA*tB;
		cnsts->pre_touch[1][0] = cnsts->pre_touch[0][1];
		cnsts->pre_touch[1][1] = tB;
		
		cnsts->touchtrunc = 1; //Uses Touch only at short range for efficiency
		
		//Calculate Touch truncation distance
		if (cnsts->touchtrunc == 1)
		{
			//Determine maximum and minimum dimensions of oids
			//Replace with stock minimum and maximum functions soon - LEJ 05/02/2009
			//A and B extrema
			double maA = maxd(cnsts->r0A,3); //max distance for A
			double miA = mind(cnsts->r0A,3);
			double maB = maxd(cnsts->r0B,3);
			double miB = mind(cnsts->r0B,3);
			
			double maT, miT;
		
			//Min and max for whole system, correcting for excluded particle types
			if (cnsts->num_A == 0)
			{
				maT = maB;
				miT = miB;
			}
			else if (cnsts->num_B == 0)
			{
				maT = maA;
				miT = miA;
			}
			else
			{
				maT = max(maA,maB);
				miT = min(miA,miB);
			}
			
			cnsts->nndmax = 2.0*nnscale*maT;
		
			//Set Touch cutoff based on eccentricity of ellipses. Cutoff will get larger as ellipses become more eccentric.
			double trs = 1.5; //Touch ratio scaling factor. Adjust only if preformance is poor.
			cnsts->tcut = 2.0*trs*maT*maT/miT;
			if (cnsts->tcut > cnsts->ljcut)
			{
				cnsts->tcut = cnsts->ljcut;
				printf("Warning: Calculated Touch truncation radius is larger than LJ cutoff. Using Touch for all LJ interactions.\n");
			}
		}
	}
	else
	{
		//Define nearest neighbor shell maximum
		cnsts->nndmax = 2.0*nnscale*max(cnsts->r0A[0],cnsts->r0B[0]);
	}
	
	//Check for invalid solvent parameters
	if (cnsts->vdwtype == 2 && cnsts->Solvent[0] == 0)
	{
		printf("Solvent[0] cannot be 0 if implicit solvent is used.\n");
		exit(1);
	}
	else if (cnsts->vdwtype == 1 || cnsts->vdwtype == 4)
	{
		double ljratio = (double)(cnsts->LJexp[1])/(double)(cnsts->LJexp[0]);
		if (ljratio != 2)
		{
			printf("Simple LJ potential requires repulsive exponent to be twice attractive component (e.g. 6 and 12)\n");
			exit(1);
		}
	}	
}

//Configure electrostatics (including reaction field)
void setup_electrostatics()
{
	Cnsts* cnsts = GetCnsts();
	
	//Set up electrostatic cutoff
	double bltemp;
	if (cnsts->latticetype < 3)
	{
		bltemp = cnsts->boxlength[0];
		if (cnsts->boxlength[1] < bltemp) bltemp = cnsts->boxlength[1];
		if (cnsts->boxlength[2] < bltemp) bltemp = cnsts->boxlength[2];
		bltemp /= 2.0;
	}
	else if (cnsts->latticetype > 2)
	{
		bltemp = cnsts->spherecylr;
		if (cnsts->latticetype == 4)
		{
			if (cnsts->boxlength[2] < bltemp/2.0) bltemp = cnsts->boxlength[2]/2.0;
		}
	}
	if (cnsts->escut > bltemp)
	{
		cnsts->escut = bltemp;
		printf("Warning: Electrostatic cutoff is too large. Resizing to half of the box length (%f)\n", bltemp);
	}
	
	//Check geometry of simulation cell
	if (cnsts->epsilon != cnsts->n2 || cnsts->scrf == 1)
	{
		if (cnsts->latticetype == 3) cnsts->rfcut = cnsts->spherecylr;
			else if (cnsts->cut == 1) cnsts->rfcut = cnsts->escut;
				else
				{
					printf("Reaction field requires a spherical cutoff or simulation cell!\n");
					exit(1);
				}
	}
	else cnsts->rfcut = 1.0; //this can be any nonzero number; it just gets multiplied by 0 - LEJ 12/22
		
	
	//Check polarizable medium dielectric constant n2
	if (abs(cnsts->n2) > 0.0) cnsts->in2 = 1.00/cnsts->n2;
	else
	{
		printf("Cannot define cavity dielectric as zero!\n");
		exit(1);
	}
	cnsts->mucorr = 1.00;
	if (cnsts->n2 != 1.00)
	{
		printf("Polarizable medium active.\n");
		if (cnsts->polarize == 1)
		{
			cnsts->mucorr *= (cnsts->n2 + 2.0)/3.0;
			printf("Onsager polarizability active.\n");
		}
	}
	
	//Check poling field if present
	if (cnsts->Efield != 0)
	{
		if (cnsts->realfield == 1)
		{
			printf("Poling field defined at constant voltage.\n");
		}
		else printf("Poling field defined as field inside cavity.\n");
	}
}

//Check that the box is valid and calculate its volume
void check_box()
{
	Cnsts* cnsts = GetCnsts();
	
	//Check to make sure PBCs are being used properly
	if (cnsts->latticetype == 3 && (cnsts->PBCs[0] == 1 || cnsts->PBCs[1] == 1 || cnsts->PBCs[2] == 1))
	{
		printf("PBCs are not yet supported for spherical simulation cells.\n");
		exit(1);
	}
	else if (cnsts->latticetype == 4 && (cnsts->PBCs[0] == 1 || cnsts->PBCs[1] == 1))
	{
		printf("PBCs are currently only supported on the z-axis for cylindrical simulation cells.\n");
		exit(1);
	}
	else if (cnsts->PBCs[0] == 0 && cnsts->PBCs[1] == 0 && cnsts->PBCs[2] == 0 && cnsts->latticetype < 3)
	{
		printf("Warning: PBCs are off! Simulating a finite cube of liquid.\n");
	}
	
	//Calculate volume for sphere, cylinder, and rectangular boxes, respectively;
	if (cnsts->latticetype == 3) cnsts->V = (4.0/3.0)*pi*qpwr(cnsts->spherecylr,3);
	else if (cnsts->latticetype == 4) cnsts->V = 2.0*pi*qpwr(cnsts->spherecylr,2)*cnsts->boxlength[2];
	else cnsts->V=cnsts->boxlength[0]*cnsts->boxlength[1]*cnsts->boxlength[2];
	
	//Check to make sure box has finite volume
	if (cnsts->V <=0.0)
	{
		printf("Box volume must be positive. Check lin_dim.");
		exit(1);
	}
}

//Check to make sure number of particles, simulation length, etc, are valid
void final_validation()
{
	Cnsts* cnsts = GetCnsts();
	
	//Make sure the simulation is worth running/runnable
	if (cnsts->n_oids == 0)
	{
		printf("The simulation doesn't contain any particles yet!\n");
		exit(1);
	}
	else if (cnsts->n_oids != (cnsts->num_A+cnsts->num_B))
	{
		printf("n_oids must equal num_A + num_B until an arbitrary number of types are allowed.\n");
		exit(1);
	}
	else if (cnsts->nndist <= 0.0)
	{
		printf("Box size may not be zero or negative!");
		exit(1);
	}
	else if (cnsts->n_oids > (cnsts->lin_dim[0]*cnsts->lin_dim[1]*cnsts->lin_dim[2]))
	{
		printf("Too many oids to fit in box!\n");
		exit(1);
	}
	else if (cnsts->n_oids < (cnsts->lin_dim[0]*cnsts->lin_dim[1]*cnsts->lin_dim[2]))
	{
		printf("The code does not currently work with empty spaces in the box.\n");
		exit(1);
	}
	else if (cnsts->stillmembersA == 1 && cnsts->stillmembersB == 1)
	{
		printf("Running a simulation where no particles can rotate or translate is rather pointless.\n");
		exit(1);
	}
	else if (cnsts->averageA == 0 && cnsts->averageB == 0)
	{
		printf("Can't calculate averages if nothing is selected for averaging.\n");
		exit(1);
	}
	else if ((cnsts->r0A[0] <= 0.0) || (cnsts->r0A[1] <= 0.0) || (cnsts->r0A[2] <= 0.0))
	{
		printf("Invalid dimension for A-type oids. All dimensions must be positive.\n");
		exit(1);
	}
	else if ((cnsts->r0B[0] <= 0.0) || (cnsts->r0B[1] <= 0.0) || (cnsts->r0B[2] <= 0.0))
	{
		printf("Invalid dimension for B-type oids. All dimensions must be positive.\n");
		exit(1);
	}	
	else if ((cnsts->vdwtype) != 4 && (cnsts->vdwtype != 5))
	{
		if (((cnsts->r0A[0] != cnsts->r0A[1]) || (cnsts->r0A[0] != cnsts->r0A[2]) || (cnsts->r0A[1] != cnsts->r0A[2])) && cnsts->num_A != 0)
		{
			printf("A-type oids are non-spherical and require Touch (vdw type 4 or 5)\n");
			exit(1);
		}
		
		if (((cnsts->r0B[0] != cnsts->r0B[1]) || (cnsts->r0B[0] != cnsts->r0B[2]) || (cnsts->r0B[1] != cnsts->r0B[2])) && cnsts->num_B != 0)
		{
			printf("B-type oids are non-spherical and require Touch (vdw type 4 or 5)\n");
			exit(1);
		}
	}
	else if (cnsts->anneal == 1 && ((cnsts->steps - cnsts->avg_steps) < cnsts->randsteps*2))
	{
		printf("Temperature is not stable during averaging period.\n");
	}
}

//A couple of basic checks for the molecules in the box — can probably be merged with other functions.
void check_oids(Molecule** BeOids)
{
	
	//Determine the number of particles actually averaged over
	Cnsts* cnsts = GetCnsts();
	int to_average = 0;
	
	for (int i = 0; i < cnsts->n_oids; i++)
	{
		if (BeOids[i]->averageme == 1) to_average += 1;
	}
	
	cnsts->n_avg = to_average;	
	
	//Warn if randomizing based on initial coordinates file.9
	if (cnsts->loader == 3 && cnsts->randsteps != 0)
	{
		printf("Warning: Initial configuration loaded from file '%s' will be randomized for %li steps.\n",cnsts->filein.c_str(), cnsts->randsteps);
	}
}

//Input file checker
void check_infile()
{
	Cnsts* cnsts = GetCnsts();
	
	//Allocate memory
	int i;
	double r0temp;
	char isend;
	char xyz[3] = {'x','y','z'};
	
	//Generate filename
	string cfn = cnsts->filein + "1chk" + cnsts->chkpt + "coords.dat";
	fstream cfile;
	
	//Check if file is used
	if (cnsts->loader == 3)
	{
		printf("Checking input file '%s'...\n",cfn.c_str());
		
		//Make sure filename is given
		if (cnsts->filein.length() == 0)
		{
			printf("No input file specified! Please specify an input file.");
			exit(4);
		}
		
		//Make sure file exists
		cfile.open(cfn.c_str(),fstream::in);
		
		//Check stream integrity
		if (cfile.good() != 1)
		{
			printf("Error opening file '%s'\n", cfn.c_str());
			exit(4);
		}
		
		//Force molecular sizes to be consistent with input file
		advanceline(cfile, 3);

		//A-type particles. Will need to fix when arbitrary #types.
		gettdint(cfile);
		for (i = 0; i < 3; i++)
		{
			r0temp = gettddoub(cfile);
			if (cnsts->r0A[i] != r0temp)
			{
				printf("Inconsistent r0A[%c] value. Using value of %f from file.\n",xyz[i],r0temp);
				cnsts->r0A[i] = r0temp;
			}
		}
		
		//B-type particles. Will need to fix when arbitrary #types.
		gettdint(cfile);
		for (i = 0; i < 3; i++)
		{
			r0temp = gettddoub(cfile);
			if (cnsts->r0B[i] != r0temp)
			{
				printf("Inconsistent r0B[%c] value. Using value of %f from file.\n",xyz[i],r0temp);
				cnsts->r0B[i] = r0temp;
			}
		}
		
		//Make sure only two molecule types are present (move this later)
		isend = cfile.peek();
		if (isend != '\n')
		{
			printf("Invalid coordinate file format. Make sure you only have two particle types.\n");
			exit(4);
		}
	}
	else if (cnsts->filein.length()!=0) printf("Starting simulation from lattice. Input file '%s' will be ignored.\n",cfn.c_str());

	//Clean up
	cfile.close();
}

//Summary file generator
//Generate machine-readable summary file. This no longer requires a batch job. LEJ 02/17/2010
void open_summary()
{
	
	Cnsts* cnsts = GetCnsts();
	
	time_t rawtime;
	struct tm * start_time;
	time (&rawtime);
	start_time = localtime (&rawtime);
	
	//Calculate corrected dipole moments, density, and reduced dipole density y
	double mzA = cnsts->muzA*cnsts->mucorr;
	double mzB = cnsts->muzB*cnsts->mucorr;
	double rho = (double)(cnsts->n_oids)/cnsts->V;
	double y = 4*pi*((double)(cnsts->num_A)*mzA*mzA + (double)(cnsts->num_B)*mzB*mzB)/(9*cnsts->V*cnsts->kT);
	
	//Generate summary file header with molecule and simulation info needed for post-processing
	string summfn = cnsts->fileout + "summary.dat";
	FILE* summ = fopen(summfn.c_str(), "wa");
	fprintf(summ, "Summary file opened at %s \n", asctime(start_time));
	fprintf(summ, "#SIMULATION SUMMARY\n\n");
	fprintf(summ, "NFILES	%i\n", cnsts->runs);
	fprintf(summ, "MOL1	%i	%f	%f\n", cnsts->num_A, cnsts->muzA, mzA);
	fprintf(summ, "MOL2	%i	%f	%f\n", cnsts->num_B, cnsts->muzB, mzB);
	fprintf(summ, "RHO	%f\n", rho);
	fprintf(summ, "T	%f\n", cnsts->T);
	fprintf(summ, "N2	%f\n", cnsts->n2);
	fprintf(summ, "Y	%f\n", y);
	fclose(summ);
}



